"""Base classes for plugin functionality. Subclass the appropriate base class
and register it in the appropriate backend's `add_..._plugin` method."""
import importlib
from typing import NotRequired, TypedDict, TypeVar

from redipy.graph.expr import CallObj, ExprObj, JSONType
from redipy.memory.state import Machine


ArgcSpec = TypedDict('ArgcSpec', {
    "count": int,
    "at_least": NotRequired[bool],
    "at_most": NotRequired[int],
})
"""Specification of number of function arguments."""


class NamedFunction:  # pylint: disable=too-few-public-methods
    """Base class for named functions."""
    @staticmethod
    def name() -> str:
        """
        The name of the function.

        Returns:
            str: The name.
        """
        raise NotImplementedError()


class GeneralFunction(NamedFunction):
    """Base class for general functions."""
    @staticmethod
    def argc() -> ArgcSpec:
        """
        The specification of the number of arguments of the function.

        Returns:
            ArgcSpec: The specification of arguments for the function.
        """
        raise NotImplementedError()


class LocalRedisFunction(GeneralFunction):
    """A redis function for the memory backend."""
    @staticmethod
    def call(sm: Machine, key: str, args: list[JSONType]) -> JSONType:
        """
        Executes the redis function for the memory backend.

        Args:
            sm (Machine): The memory redis state.

            key (str): The key.

            args (list[JSONType]): The rest of the arguments.

        Returns:
            JSONType: The result of the redis function.
        """
        raise NotImplementedError()


class LocalGeneralFunction(GeneralFunction):
    """A general script function for the memory backend."""
    @staticmethod
    def call(args: list[JSONType]) -> JSONType:
        """
        Executes a general function for the memory backend.

        Args:
            args (list[JSONType]): The arguments.

        Returns:
            JSONType: The result of the general function.
        """
        raise NotImplementedError()


HELPER_PKG = "redipy"
"""The name of the helper package in lua."""


class LuaPatch:
    """Base class for lua patches."""
    def __init__(self, name: str) -> None:
        self._name = name

    def name(self) -> str:
        """
        The name of the patched function.

        Returns:
            str: One of the values of names.
        """
        return self._name

    @staticmethod
    def names() -> set[str]:
        """
        All names the patch is applied to.

        Returns:
            set[str]: All function names the patch is applied to.
        """
        raise NotImplementedError()

    def helper_pkg(self) -> str:
        """
        The name of the lua helper package.

        Returns:
            str: The lua helper package.
        """
        return HELPER_PKG


class LuaRedisPatch(LuaPatch):
    """Patches a lua redis function call."""
    def patch(
            self,
            name: str,
            expr: CallObj,
            args: list[ExprObj],
            *,
            is_expr_stmt: bool) -> ExprObj:
        """
        Applies the patch on the expression graph for the given redis call.

        Args:
            name (str): The redis function name.

            expr (CallObj): The function call.

            args (list[ExprObj]): The arguments of the function call.

            is_expr_stmt (bool): Whether the call is a lua statement. Most
            patches should not be applied in this context as lua doesn't allow
            expressions as statements (whereas function calls can be both).
            In most cases add the following at the beginning of the patch
            implementation:
            ```
            if is_expr_stmt:
                return expr
            ```

        Returns:
            ExprObj: The patched expression.
        """
        raise NotImplementedError()


class LuaGeneralPatch(LuaPatch):
    """Patches a general lua function call."""
    def patch(self, expr: CallObj, *, is_expr_stmt: bool) -> ExprObj:
        """
        Applies the patch on the expression graph for the given function call.

        Args:
            expr (CallObj): The function call.

            is_expr_stmt (bool): Whether the call is a lua statement. Most
            patches should not be applied in this context as lua doesn't allow
            expressions as statements (whereas function calls can be both).
            In most cases add the following at the beginning of the patch
            implementation:
            ```
            if is_expr_stmt:
                return expr
            ```

        Returns:
            ExprObj: The patched expression.
        """
        raise NotImplementedError()


class HelperFunction(NamedFunction):
    """Defines a lua helper function."""
    @staticmethod
    def args() -> str:
        """
        The literal argument line.

        Returns:
            str: The argument line as string (the content between parentheses).
        """
        raise NotImplementedError()

    @staticmethod
    def body() -> str:
        """
        The literal function body.

        Returns:
            str: The function body as multi-line string. Use 4 space
            indentation. The content will be deindented and the indentation
            will be adjusted to lua's 2 spaces when loading the function.
        """
        raise NotImplementedError()


T = TypeVar('T', bound=NamedFunction)


def add_plugin(
        module: str,
        target: dict[str, T],
        clazz: type[T],
        disallowed: set[str] | None = None) -> None:
    """
    Adds eligible elements of the given module as plugins to the target.

    Args:
        module (str): The module to read the plugins from.

        target (dict[str, T]): The destination to store the plugins.

        clazz (type[T]): Only elements of this class are loaded as plugins.

        disallowed (set[str] | None, optional): A set of invalid names.
        If a plugin attempts to use one of those names an error is raised.
        Defaults to None.

    Raises:
        RuntimeError: If a plugin cannot be loaded.
    """
    if disallowed is None:
        disallowed = set()
    mod = importlib.import_module(module)
    candidates = [
        cls
        for cls in mod.__dict__.values()
        if isinstance(cls, type)
        and cls.__module__ == module
        and issubclass(cls, clazz)
    ]
    for rfun in candidates:
        rname = rfun.name()
        if rname in disallowed:
            raise RuntimeError(
                f"function name {rname} is not allowed")
        if rname in target:
            raise RuntimeError(
                f"duplicate function definition: {rname}")
        target[rname] = rfun()


U = TypeVar('U', bound=LuaPatch)


def add_patch_plugin(
        module: str,
        target: dict[str, U],
        clazz: type[U],
        disallowed: set[str] | None = None) -> None:
    """
    Adds eligible elements of the given module as plugins to the target.
    The elements must be lua patches.

    Args:
        module (str): The module to read the plugins from.

        target (dict[str, U]): The destination to store the plugins.

        clazz (type[U]): Only elements of this class are loaded as plugins.

        disallowed (set[str] | None, optional): A set of invalid names.
        If a plugin attempts to use one of those names an error is raised.
        Defaults to None.

    Raises:
        RuntimeError: If a plugin cannot be loaded.
    """
    if disallowed is None:
        disallowed = set()
    mod = importlib.import_module(module)
    candidates = [
        cls
        for cls in mod.__dict__.values()
        if isinstance(cls, type)
        and cls.__module__ == module
        and issubclass(cls, clazz)
    ]
    for pfun in candidates:
        for pname in pfun.names():
            if pname in disallowed:
                raise RuntimeError(
                    f"patch name {pname} is not allowed")
            if pname in target:
                raise RuntimeError(
                    f"duplicate patch definition: {pname}")
            target[pname] = pfun(pname)



"""Tests redis values."""
import json
import time
from collections.abc import Callable
from test.util import get_setup, get_test_config

import pytest

from redipy.api import RSM_EXISTS, RSM_MISSING
from redipy.graph.expr import JSONType
from redipy.memory.local import LocalBackend
from redipy.memory.rt import LocalRuntime
from redipy.redis.conn import RedisConnection
from redipy.redis.lua import LuaBackend
from redipy.symbolic.expr import Expr
from redipy.symbolic.fun import ToNum, ToStr, TypeStr
from redipy.symbolic.rvar import RedisVar
from redipy.symbolic.seq import FnContext
from redipy.util import code_fmt, lua_fmt


LUA_SCRIPT = lua_fmt("""
--[[ KEYV
k
]]
--[[ ARGV
a
]]
local arg_0 = cjson.decode(ARGV[1])  -- a
local key_0 = (KEYS[1])  -- k
if (tonumber((redis.call("get", key_0) or 0)) <= arg_0) then
    redis.call("set", key_0, arg_0)
end
local var_0 = (redis.call("get", key_0) or nil)
if (var_0 ~= nil) then
    var_0 = tonumber(var_0)
end
return cjson.encode(var_0)
""")


RUN_TESTS: list[tuple[str, JSONType, JSONType]] = [
    ("foo", 1, 1),
    ("foo", 3, 3),
    ("foo", 2, 3),
    ("bar", 5, 5),
    ("bar", 2, 5),
    ("foo", 4, 4),
    ("bar", 5, 5),
    ("foo", 8, 8),
    ("foo", 9, 9),
    ("neg", -1, None),
    ("neg", 0, 0),
    ("neg", 1, 1),
    ("neg", 0, 1),
    ("bar", 2, 5),
    ("pos", 3, 5),
    ("bar", 7, 7),
    ("pos", 6, 6),
]


def test_rvar() -> None:
    """Tests redis values with an additional explicit script."""
    ctx = FnContext()
    a = ctx.add_arg("a")
    k = RedisVar(ctx.add_key("k"))
    b_then, _ = ctx.if_(ToNum(k.get(no_adjust=True).or_(0)).le_(a))
    b_then.add(k.set(a))
    res = ctx.add_local(k.get())
    r_then, _ = ctx.if_(res.ne_(None))
    r_then.add(res.assign(ToNum(res)))
    ctx.set_return_value(res)

    compiled = ctx.compile()
    print(json.dumps(compiled, indent=2, sort_keys=True))

    lcl = LocalBackend()
    lcl_code = lcl.translate(compiled)
    lrt = LocalRuntime()
    run_lcl = lcl.create_executable(lcl_code, lrt)

    lrt.set("pos", "5")
    for (k_in, a_in, expect_out) in RUN_TESTS:
        is_out = run_lcl(keys={"k": k_in}, args={"a": a_in})
        assert is_out == expect_out
    assert lrt.get("foo") == "9"
    assert lrt.get("bar") == "7"
    assert lrt.get("neg") == "1"
    assert lrt.get("pos") == "6"

    lua = LuaBackend()
    lua_code = lua.translate(compiled)
    assert code_fmt(lua_code) == LUA_SCRIPT

    rrt = RedisConnection("test_rvar", cfg=get_test_config())
    run_redis = lua.create_executable(lua_code, rrt)

    rrt.set("pos", "5")
    for (k_in, a_in, expect_out) in RUN_TESTS:
        is_out = run_redis(keys={"k": k_in}, args={"a": a_in})
        assert is_out == expect_out
    assert rrt.get("foo") == "9"
    assert rrt.get("bar") == "7"
    assert rrt.get("neg") == "1"
    assert rrt.get("pos") == "6"


@pytest.mark.parametrize("rt_lua", [False, True])
def test_set_ext_args(rt_lua: bool) -> None:
    """
    Tests key expiration.

    Args:
        rt_lua (bool): Whether to use the redis or memory runtime.
    """
    rt = get_setup(
        "test_set_ext_args", rt_lua, no_compile_hook=True)

    # FIXME: test expire_timestamp

    def fun_check(
            key: str,
            name: str,
            expr: Callable[[RedisVar, Expr], Expr],
            type_str: str,
            value_str: str,
            lua_snippet: str) -> None:
        lua_script = f"""
            --[[ KEYV
            in
            ]]
            --[[ ARGV
            in
            ]]
            local arg_0 = cjson.decode(ARGV[1])  -- in
            local key_0 = (KEYS[1])  -- in
            local var_0 = {lua_snippet}
            local var_1 = cjson.decode("[]")
            var_1[#var_1 + 1] = ""
            var_1[#var_1 + 1] = ""
            var_1[0 + 1] = type(var_0)
            var_1[1 + 1] = tostring(var_0)
            return cjson.encode(var_1)
        """

        def code_hook(code: list[str]) -> None:
            code_str = code_fmt(code)
            assert code_str == lua_fmt(lua_script)
            print(code_str)

        if rt_lua:
            rt.set_code_hook(code_hook)

        ctx = FnContext()
        arg_in = ctx.add_arg("in")
        rvar = RedisVar(ctx.add_key("in"))
        res_val = ctx.add_local(expr(rvar, arg_in))
        res_arr = ctx.add_local([])
        ctx.add(res_arr.set_at(res_arr.len_(), ""))
        ctx.add(res_arr.set_at(res_arr.len_(), ""))
        ctx.add(res_arr.set_at(0, TypeStr(res_val)))
        ctx.add(res_arr.set_at(1, ToStr(res_val)))
        ctx.set_return_value(res_arr)

        exec_fun = rt.register_script(ctx)
        res = exec_fun(keys={"in": key}, args={"in": name})
        assert isinstance(res, list)
        assert len(res) == 2
        assert res[0] == type_str
        assert res[1] == value_str

    assert rt.get("bar") is None
    assert rt.set("bar", "a") is True
    assert rt.get("bar") == "a"
    assert rt.set("bar", "c", mode=RSM_MISSING) is False
    assert rt.get("bar") == "a"
    assert rt.set("bar", "b", return_previous=True, expire_in=0.1) == "a"
    assert rt.get("bar") == "b"
    assert rt.set("bar", "c", mode=RSM_MISSING, keep_ttl=True) is False
    assert rt.get("bar") == "b"
    assert rt.set("bar", "e", mode=RSM_EXISTS, keep_ttl=True) is True
    assert rt.get("bar") == "e"

    with rt.pipeline() as pipe:
        pipe.get("baz")  # 0
        pipe.set("baz", "-", mode=RSM_EXISTS)  # 1
        pipe.get("baz")  # 2
        pipe.set("baz", "a")  # 3
        pipe.get("baz")  # 4
        pipe.set("baz", "b", mode=RSM_EXISTS)  # 5
        pipe.get("baz")  # 6
        pipe.set("baz", "c", mode=RSM_MISSING)  # 7
        pipe.get("baz")  # 8
        pipe.set("baz", "d", return_previous=True, expire_in=0.1)  # 9
        pipe.get("baz")  # 10
        pipe_res = pipe.execute()
    assert pipe_res[0] is None
    assert pipe_res[1] is False
    assert pipe_res[2] is None
    assert pipe_res[3] is True
    assert pipe_res[4] == "a"
    assert pipe_res[5] is True
    assert pipe_res[6] == "b"
    assert pipe_res[7] is False
    assert pipe_res[8] == "b"
    assert pipe_res[9] == "b"
    assert pipe_res[10] == "d"

    assert rt.get("foo") is None
    fun_check(
        "foo",
        "a",
        lambda rvar, name: rvar.set(name),
        "boolean",
        "true",
        "(redis.call(\"set\", key_0, arg_0) ~= false)")
    assert rt.get("foo") == "a"
    fun_check(
        "foo",
        "c",
        lambda rvar, name: rvar.set(name, mode=RSM_MISSING),
        "boolean",
        "false",
        "(redis.call(\"set\", key_0, arg_0, \"NX\") ~= false)")
    assert rt.get("foo") == "a"
    fun_check(
        "foo",
        "b",
        lambda rvar, name: rvar.set(name, return_previous=True, expire_in=0.1),
        "string",
        "a",
        "redis.call(\"set\", key_0, arg_0, \"GET\", \"PX\", 100)")
    assert rt.get("foo") == "b"
    fun_check(
        "foo",
        "c",
        lambda rvar, name: rvar.set(name, mode=RSM_MISSING, keep_ttl=True),
        "boolean",
        "false",
        "(redis.call(\"set\", key_0, arg_0, \"NX\", \"KEEPTTL\") ~= false)")
    assert rt.get("foo") == "b"
    fun_check(
        "foo",
        "e",
        lambda rvar, name: rvar.set(name, mode=RSM_EXISTS, keep_ttl=True),
        "boolean",
        "true",
        "(redis.call(\"set\", key_0, arg_0, \"XX\", \"KEEPTTL\") ~= false)")
    assert rt.get("foo") == "e"

    time.sleep(0.1)

    assert rt.set("bar", "d", mode=RSM_MISSING) is True
    assert rt.get("bar") == "d"

    assert rt.get("baz") is None

    fun_check(
        "foo",
        "d",
        lambda rvar, name: rvar.set(name, mode=RSM_MISSING),
        "boolean",
        "true",
        "(redis.call(\"set\", key_0, arg_0, \"NX\") ~= false)")
    assert rt.get("foo") == "d"



"""Module for patching lua redis function calls."""
from redipy.graph.expr import CallObj, ExprObj, find_literal, LiteralValObj
from redipy.plugin import LuaRedisPatch


class RSetPatch(LuaRedisPatch):
    """Converts the output of SET into a proper boolean value."""
    @staticmethod
    def names() -> set[str]:
        return {"set"}

    def patch(
            self,
            name: str,
            expr: CallObj,
            args: list[ExprObj],
            *,
            is_expr_stmt: bool) -> ExprObj:
        if is_expr_stmt:
            return expr
        if find_literal(
                args[1:], "GET", vtype="str", no_case=True) is not None:
            return expr
        return {
            "kind": "binary",
            "op": "ne",
            "left": expr,
            "right": {
                "kind": "val",
                "type": "bool",
                "value": False,
            },
        }


class RGetPatch(LuaRedisPatch):
    """Ensures GET-like operations return None (nil) if the key or value is
    missing."""
    @staticmethod
    def names() -> set[str]:
        return {"get", "lpop", "rpop", "hget"}

    def patch(
            self,
            name: str,
            expr: CallObj,
            args: list[ExprObj],
            *,
            is_expr_stmt: bool) -> ExprObj:
        if is_expr_stmt:
            return expr
        # check if 2nd argument (count) exists for lpop or rpop
        if len(args) > 1 and name in ["lpop", "rpop"]:
            return expr
        return {
            "kind": "binary",
            "op": "or",
            "left": expr,
            "right": {
                "kind": "val",
                "type": "none",
                "value": None,
            },
        }


class RSortedPopPatch(LuaRedisPatch):
    """Converts the output of sorted set list functions into pair lists."""
    @staticmethod
    def names() -> set[str]:
        return {"zpopmax", "zpopmin"}

    def patch(
            self,
            name: str,
            expr: CallObj,
            args: list[ExprObj],
            *,
            is_expr_stmt: bool) -> ExprObj:
        if is_expr_stmt:
            return expr
        return {
            "kind": "call",
            "name": f"{self.helper_pkg()}.pairlist_scores",
            "args": [expr],
            "no_adjust": False,
        }


class RIncrByPatch(LuaRedisPatch):
    """Uses the float variant for INCRBY-like functions and convert the result
    to a number."""
    @staticmethod
    def names() -> set[str]:
        return {"incrby", "hincrby"}

    def patch(
            self,
            name: str,
            expr: CallObj,
            args: list[ExprObj],
            *,
            is_expr_stmt: bool) -> ExprObj:
        name = f"{name}float"
        literal: LiteralValObj = {
            "kind": "val",
            "type": "str",
            "value": name,
        }
        expr["args"][0] = literal
        if is_expr_stmt:
            return expr
        return {
            "kind": "call",
            "name": "tonumber",
            "args": [expr],
            "no_adjust": False,
        }


class RHashGetSomePatch(LuaRedisPatch):
    """Uses keys and values to build a dictionary."""
    @staticmethod
    def names() -> set[str]:
        return {"hmget"}

    def patch(
            self,
            name: str,
            expr: CallObj,
            args: list[ExprObj],
            *,
            is_expr_stmt: bool) -> ExprObj:
        if is_expr_stmt:
            return expr
        return {
            "kind": "call",
            "name": f"{self.helper_pkg()}.keyval_dict",
            "args": [expr, *args[1:]],
            "no_adjust": False,
        }


class RHashGetAllPatch(LuaRedisPatch):
    """Converts an alternating key value list into a dictionary."""
    @staticmethod
    def names() -> set[str]:
        return {"hgetall"}

    def patch(
            self,
            name: str,
            expr: CallObj,
            args: list[ExprObj],
            *,
            is_expr_stmt: bool) -> ExprObj:
        if is_expr_stmt:
            return expr
        return {
            "kind": "call",
            "name": f"{self.helper_pkg()}.pairlist_dict",
            "args": [expr],
            "no_adjust": False,
        }



"""Access to redis sorted sets."""
from redipy.symbolic.expr import Expr, MixedType
from redipy.symbolic.fun import RedisObj


class RedisSortedSet(RedisObj):
    """A redis sorted set."""
    def add(self, score: MixedType, value: MixedType) -> Expr:
        """
        Adds an element to the sorted set.

        Args:
            score (MixedType): The score.
            value (MixedType): The value.

        Returns:
            Expr: The expression.
        """
        # FIXME add all arguments
        return self.redis_fn("zadd", score, value)

    def pop_max(self, count: MixedType = None) -> Expr:
        """
        Pops maximum values from the sorted set.

        Args:
            count (MixedType, optional): The number of values to pop. Defaults
            to one (None).

        Returns:
            Expr: The expression.
        """
        if count is None:
            return self.redis_fn("zpopmax")
        return self.redis_fn("zpopmax", count)

    def pop_min(self, count: MixedType = None) -> Expr:
        """
        Pops minimum values from the sorted set.

        Args:
            count (MixedType, optional): The number of values to pop. Defaults
            to one (None).

        Returns:
            Expr: The expression.
        """
        if count is None:
            return self.redis_fn("zpopmin")
        return self.redis_fn("zpopmin", count)

    def range(self, start: MixedType, stop: MixedType) -> Expr:
        """
        Returns a range of member names.

        Args:
            start (MixedType): The start index.

            stop (MixedType): The stop index (inclusive).

        Returns:
            Expr: The expression.
        """
        # FIXME add all arguments
        return self.redis_fn("zrange", start, stop)

    def card(self) -> Expr:
        """
        Computes the cardinality of the sorted set.

        Returns:
            Expr: The expression.
        """
        return self.redis_fn("zcard")

    # FIXME implement ZRANGE



"""Access to redis lists."""
from redipy.symbolic.expr import Expr, MixedType
from redipy.symbolic.fun import RedisObj


class RedisList(RedisObj):
    """A redis list."""
    def lpush(self, *values: MixedType) -> Expr:
        """
        Pushes values to the left end of the list.

        Args:
            *values (MixedType): The values to push.

        Returns:
            Expr: The expression.
        """
        return self.redis_fn("lpush", *values)

    def rpush(self, *values: MixedType) -> Expr:
        """
        Pushes values to the right end of the list.

        Args:
            *values (MixedType): The values to push.

        Returns:
            Expr: The expression.
        """
        return self.redis_fn("rpush", *values)

    def lpop(
            self, count: MixedType = None, *, no_adjust: bool = False) -> Expr:
        """
        Pops values from the left end of the list.

        Args:
            count (MixedType, optional): The number of values to pop.
            Defaults to one (None) returning a singleton.

            no_adjust (bool, optional): Whether to prevent patching the
            function call. This should not be neccessary. Defaults to False.

        Returns:
            Expr: The expression.
        """
        if count is None:
            return self.redis_fn("lpop", no_adjust=no_adjust)
        return self.redis_fn("lpop", count, no_adjust=no_adjust)

    def rpop(
            self, count: MixedType = None, *, no_adjust: bool = False) -> Expr:
        """
        Pops values from the right end of the list.

        Args:
            count (MixedType, optional): The number of values to pop.
            Defaults to one (None) returning a singleton.

            no_adjust (bool, optional): Whether to prevent patching the
            function call. This should not be neccessary. Defaults to False.

        Returns:
            Expr: The expression.
        """
        if count is None:
            return self.redis_fn("rpop", no_adjust=no_adjust)
        return self.redis_fn("rpop", count, no_adjust=no_adjust)

    def lrange(self, start: MixedType, stop: MixedType) -> Expr:
        """
        Returns a number of values from the list specified by the given range.
        Negative numbers are interpreted as index from the back of the list.
        Out of range indices are ignored, potentially returning an empty list.

        Args:
            start (MixedType): The start index.

            stop (MixedType): The stop index (inclusive).

        Returns:
            Expr: The expression.
        """
        return self.redis_fn("lrange", start, stop)

    def llen(self) -> Expr:
        """
        The length of the list.

        Returns:
            Expr: The expression.
        """
        return self.redis_fn("llen")



"""Redipy is a backend agnostic redis interface. It supports an in process
memory redis backend and a redis server connection backend. Redis functions can
be created using functionality of the `redipy.script` module. New functions or
redis functionality can be added with the help of the `redipy.plugin` module.

The most common symbols of redipy are reexported at the top level for easy
access."""
import redipy.plugin  # pylint: disable=unused-import  # noqa
import redipy.script  # pylint: disable=unused-import  # noqa
from redipy.api import (
    PipelineAPI,
    RedisAPI,
    RedisClientAPI,
    RSetMode,
    RSM_ALWAYS,
    RSM_EXISTS,
    RSM_MISSING,
)
from redipy.backend.backend import ExecFunction
from redipy.backend.runtime import Runtime
from redipy.main import Redis
from redipy.memory.rt import LocalRuntime
from redipy.redis.conn import RedisConfig, RedisConnection, RedisFactory


__all__ = [
    "ExecFunction",
    "LocalRuntime",
    "PipelineAPI",
    "plugin",
    "Redis",
    "RedisAPI",
    "RedisClientAPI",
    "RedisConfig",
    "RedisConnection",
    "RedisFactory",
    "RSetMode",
    "RSM_ALWAYS",
    "RSM_EXISTS",
    "RSM_MISSING",
    "Runtime",
    "script",
]



"""The runtime for the memory backend."""
import contextlib
import datetime
from collections.abc import Callable, Iterator
from typing import Any, Literal, overload, TypeVar

from redipy.api import PipelineAPI, RSetMode, RSM_ALWAYS
from redipy.backend.runtime import Runtime
from redipy.graph.expr import JSONType
from redipy.memory.local import Cmd, LocalBackend
from redipy.memory.state import Machine, State
from redipy.plugin import add_plugin, LocalGeneralFunction, LocalRedisFunction


T = TypeVar('T')
C = TypeVar('C', bound=Callable)


CONST: dict[str, JSONType] = {
    "redis.LOG_DEBUG": "DEBUG",
    "redis.LOG_VERBOSE": "VERBOSE",
    "redis.LOG_NOTICE": "NOTICE",
    "redis.LOG_WARNING": "WARNING",
}
"""Log severity levels."""


class LocalRuntime(Runtime[Cmd]):
    """The runtime of the memory backend."""
    def __init__(self) -> None:
        super().__init__()
        self._sm = Machine(State())
        self._rfuns: dict[str, LocalRedisFunction] = {}
        self._gfuns: dict[str, LocalGeneralFunction] = {}
        self.add_redis_function_plugin("redipy.memory.rfun")
        self.add_general_function_plugin("redipy.memory.gfun")

    def add_redis_function_plugin(self, module: str) -> None:
        """Adds all redis functions (LocalRedisFunction) defined in the
        given module.

        Args:
            module (str): The module name.
        """
        add_plugin(module, self._rfuns, LocalRedisFunction)

    def add_general_function_plugin(self, module: str) -> None:
        """Adds all general functions (LocalGeneralFunction) defined in the
        given module

        Args:
            module (str): The module name.
        """
        add_plugin(
            module,
            self._gfuns,
            LocalGeneralFunction,
            disallowed={"redis.call"})

    def get_machine(self) -> Machine:
        """
        Returns the internal memory state.

        Returns:
            Machine: The memory state.
        """
        return self._sm

    @classmethod
    def create_backend(cls) -> LocalBackend:
        return LocalBackend()

    @contextlib.contextmanager
    def pipeline(self) -> Iterator[PipelineAPI]:

        def exec_call(execute: Callable[[], list]) -> list:
            with self.lock():
                res = execute()
                state = pipe.get_state()
                self._sm.get_state().apply(state)
                state.reset()
                return res

        pipe = LocalPipeline(self, self._sm.get_state(), exec_call)
        yield pipe
        if pipe.has_pending():
            pipe.execute()

    @staticmethod
    def require_argc(
            args: list[JSONType],
            count: int,
            *,
            at_least: bool = False,
            at_most: int | None = None) -> None:
        """
        Confirms that the passed arguments conform to the argument
        specification.

        Args:
            args (list[JSONType]): The arguments to check.

            count (int): The expected number of arguments.

            at_least (bool, optional): If the number of arguments is a minimum.
            Defaults to False.

            at_most (int | None, optional): Defines an upper bound for the
            number of arguments. This argument does not define a minimum
            number of arguments (if at_least is set to True and at_most is
            set to a value the minimum number of arguments is 0).
            Defaults to None.

        Raises:
            ValueError: If the arguments don't conform the specification.
        """
        argc = len(args)
        if argc == count:
            return
        if at_most is not None and argc <= at_most:
            return
        if at_least and argc > count:
            return
        at_most_str = "" if at_most is None else f"up to {at_most} arguments "
        raise ValueError(
            "incorrect number of arguments need "
            f"{'at least' if at_least else 'exactly'} {count} "
            f"{at_most_str}got {argc}")

    def redis_fn(
            self,
            sm: Machine,
            name: str,
            args: list[JSONType]) -> JSONType:
        """
        Calls a redis function. For internal use only.

        Args:
            sm (Machine): The state of the memory runtime.

            name (str): The redis function name.

            args (list[JSONType]): The argument list.

        Raises:
            ValueError: If the function cannot be called.

        Returns:
            JSONType: The result of the function call.
        """
        key = f"{args[0]}"
        rfun = self._rfuns.get(name)
        if rfun is None:
            raise ValueError(f"unknown redis function {name}")
        argc = rfun.argc()
        count = argc["count"] + 1
        at_least = argc.get("at_least", False)
        at_most = argc.get("at_most")
        if at_most is not None:
            at_most += 1
        self.require_argc(args, count, at_least=at_least, at_most=at_most)
        return rfun.call(sm, key, args[1:])

    def call_fn(
            self, sm: Machine, name: str, args: list[JSONType]) -> JSONType:
        """
        Calls a function. For internal use only.

        Args:
            sm (Machine): The state of the memory runtime.

            name (str): The function name.

            args (list[JSONType]): The arguments.

        Raises:
            ValueError: If the function cannot be called.

        Returns:
            JSONType: The result of the function call.
        """
        if name == "redis.call":
            self.require_argc(args, 2, at_least=True)
            return self.redis_fn(sm, f"{args[0]}", args[1:])
        gfun = self._gfuns.get(name)
        if gfun is None:
            raise ValueError(f"unknown function {name}")
        argc = gfun.argc()
        count = argc["count"]
        at_least = argc.get("at_least", False)
        at_most = argc.get("at_most")
        self.require_argc(args, count, at_least=at_least, at_most=at_most)
        return gfun.call(args)

    def get_constant(self, raw: str) -> JSONType:
        """
        Returns a constant value. For internal use only.

        Args:
            raw (str): The name of the constant.

        Returns:
            JSONType: The value.
        """
        return CONST[raw]

    def exists(self, *keys: str) -> int:
        with self.lock():
            return self._sm.exists(*keys)

    def delete(self, *keys: str) -> int:
        with self.lock():
            return self._sm.delete(*keys)

    @overload
    def set(
            self,
            key: str,
            value: str,
            *,
            mode: RSetMode,
            return_previous: Literal[True],
            expire_timestamp: datetime.datetime | None,
            expire_in: float | None,
            keep_ttl: bool) -> str | None:
        ...

    @overload
    def set(
            self,
            key: str,
            value: str,
            *,
            mode: RSetMode,
            return_previous: Literal[False],
            expire_timestamp: datetime.datetime | None,
            expire_in: float | None,
            keep_ttl: bool) -> bool | None:
        ...

    @overload
    def set(
            self,
            key: str,
            value: str,
            *,
            mode: RSetMode = RSM_ALWAYS,
            return_previous: bool = False,
            expire_timestamp: datetime.datetime | None = None,
            expire_in: float | None = None,
            keep_ttl: bool = False) -> str | bool | None:
        ...

    def set(
            self,
            key: str,
            value: str,
            *,
            mode: RSetMode = RSM_ALWAYS,
            return_previous: bool = False,
            expire_timestamp: datetime.datetime | None = None,
            expire_in: float | None = None,
            keep_ttl: bool = False) -> str | bool | None:
        with self.lock():
            return self._sm.set(
                key,
                value,
                mode=mode,
                return_previous=return_previous,
                expire_timestamp=expire_timestamp,
                expire_in=expire_in,
                keep_ttl=keep_ttl)

    def get(self, key: str) -> str | None:
        with self.lock():
            return self._sm.get(key)

    def incrby(self, key: str, inc: float | int) -> float:
        with self.lock():
            return self._sm.incrby(key, inc)

    def lpush(self, key: str, *values: str) -> int:
        with self.lock():
            return self._sm.lpush(key, *values)

    def rpush(self, key: str, *values: str) -> int:
        with self.lock():
            return self._sm.rpush(key, *values)

    @overload
    def lpop(
            self,
            key: str,
            count: None = None) -> str | None:
        ...

    @overload
    def lpop(  # pylint: disable=signature-differs
            self,
            key: str,
            count: int) -> list[str] | None:
        ...

    def lpop(
            self,
            key: str,
            count: int | None = None) -> str | list[str] | None:
        with self.lock():
            return self._sm.lpop(key, count)

    @overload
    def rpop(
            self,
            key: str,
            count: None = None) -> str | None:
        ...

    @overload
    def rpop(  # pylint: disable=signature-differs
            self,
            key: str,
            count: int) -> list[str] | None:
        ...

    def rpop(
            self,
            key: str,
            count: int | None = None) -> str | list[str] | None:
        with self.lock():
            return self._sm.rpop(key, count)

    def lrange(self, key: str, start: int, stop: int) -> list[str]:
        with self.lock():
            return self._sm.lrange(key, start, stop)

    def llen(self, key: str) -> int:
        with self.lock():
            return self._sm.llen(key)

    def zadd(self, key: str, mapping: dict[str, float]) -> int:
        with self.lock():
            return self._sm.zadd(key, mapping)

    def zpop_max(
            self,
            key: str,
            count: int = 1,
            ) -> list[tuple[str, float]]:
        with self.lock():
            return self._sm.zpop_max(key, count)

    def zpop_min(
            self,
            key: str,
            count: int = 1,
            ) -> list[tuple[str, float]]:
        with self.lock():
            return self._sm.zpop_min(key, count)

    def zrange(self, key: str, start: int, stop: int) -> list[str]:
        with self.lock():
            return self._sm.zrange(key, start, stop)

    def zcard(self, key: str) -> int:
        with self.lock():
            return self._sm.zcard(key)

    def hset(self, key: str, mapping: dict[str, str]) -> int:
        with self.lock():
            return self._sm.hset(key, mapping)

    def hdel(self, key: str, *fields: str) -> int:
        with self.lock():
            return self._sm.hdel(key, *fields)

    def hget(self, key: str, field: str) -> str | None:
        with self.lock():
            return self._sm.hget(key, field)

    def hmget(self, key: str, *fields: str) -> dict[str, str | None]:
        with self.lock():
            return self._sm.hmget(key, *fields)

    def hincrby(self, key: str, field: str, inc: float | int) -> float:
        with self.lock():
            return self._sm.hincrby(key, field, inc)

    def hkeys(self, key: str) -> list[str]:
        with self.lock():
            return self._sm.hkeys(key)

    def hvals(self, key: str) -> list[str]:
        with self.lock():
            return self._sm.hvals(key)

    def hgetall(self, key: str) -> dict[str, str]:
        with self.lock():
            return self._sm.hgetall(key)

    def __str__(self) -> str:
        return f"{self.__class__.__name__}[{self._sm.get_state()}]"

    def __repr__(self) -> str:
        return self.__str__()


class LocalPipeline(PipelineAPI):
    """A pipeline for the memory backend runtime."""
    def __init__(
            self,
            rt: LocalRuntime,
            parent: State,
            exec_call: Callable[[Callable[[], list]], list]) -> None:
        """
        Creates a new pipeline. Do not manually create a pipeline. Use the
        `pipeline` function of the runtime instead.

        Args:
            rt (LocalRuntime): The memory backend runtime.

            parent (State): The memory state.

            exec_call (Callable[[Callable[[], list]], list]): Function that is
            called with the results of the pipeline when calling execute. This
            can be used to finalize the results before returning them.
        """
        super().__init__()
        self._rt = rt
        self._sm = Machine(State(parent))
        self._exec_call = exec_call
        self._cmd_queue: list[Callable[[], Any]] = []

    def get_runtime_tuple(self) -> tuple[LocalRuntime, Machine]:
        """
        Returns the access and state of the runtime.

        Returns:
            tuple[LocalRuntime, Machine]: The runtime object and the internal
            memory state of the pipeline.
        """
        return (self._rt, self._sm)

    def get_state(self) -> State:
        """
        Returns the raw access to the memory of the pipeline.

        Returns:
            State: The pipeline memory state.
        """
        return self._sm.get_state()

    def has_pending(self) -> bool:
        """
        Whether there are commands in the pipeline that have not been executed
        yet.

        Returns:
            bool: True if there are unexecuted commands in the pipeline.
        """
        return len(self._cmd_queue) > 0

    def execute(self) -> list:
        cmds = self._cmd_queue
        self._cmd_queue = []

        def executor() -> list:
            return [cmd() for cmd in cmds]

        return self._exec_call(executor)

    def add_cmd(self, cb: Callable[[], Any]) -> None:
        """
        Adds a command to the pipeline.

        Args:
            cb (Callable[[], Any]): The command.
        """
        self._cmd_queue.append(cb)

    def exists(self, *keys: str) -> None:
        self.add_cmd(lambda: self._sm.exists(*keys))

    def delete(self, *keys: str) -> None:
        self.add_cmd(lambda: self._sm.delete(*keys))

    def set(
            self,
            key: str,
            value: str,
            *,
            mode: RSetMode = RSM_ALWAYS,
            return_previous: bool = False,
            expire_timestamp: datetime.datetime | None = None,
            expire_in: float | None = None,
            keep_ttl: bool = False) -> None:
        self.add_cmd(lambda: self._sm.set(
            key,
            value,
            mode=mode,
            return_previous=return_previous,
            expire_timestamp=expire_timestamp,
            expire_in=expire_in,
            keep_ttl=keep_ttl))

    def get(self, key: str) -> None:
        self.add_cmd(lambda: self._sm.get(key))

    def incrby(self, key: str, inc: float | int) -> None:
        self.add_cmd(lambda: self._sm.incrby(key, inc))

    def lpush(self, key: str, *values: str) -> None:
        self.add_cmd(lambda: self._sm.lpush(key, *values))

    def rpush(self, key: str, *values: str) -> None:
        self.add_cmd(lambda: self._sm.rpush(key, *values))

    def lpop(
            self,
            key: str,
            count: int | None = None) -> None:
        self.add_cmd(lambda: self._sm.lpop(key, count))

    def rpop(
            self,
            key: str,
            count: int | None = None) -> None:
        self.add_cmd(lambda: self._sm.rpop(key, count))

    def lrange(self, key: str, start: int, stop: int) -> None:
        self.add_cmd(lambda: self._sm.lrange(key, start, stop))

    def llen(self, key: str) -> None:
        self.add_cmd(lambda: self._sm.llen(key))

    def zadd(self, key: str, mapping: dict[str, float]) -> None:
        lmap = {
            name: float(num)
            for name, num in mapping.items()
        }
        self.add_cmd(lambda: self._sm.zadd(key, lmap))

    def zpop_max(
            self,
            key: str,
            count: int = 1,
            ) -> None:
        self.add_cmd(lambda: self._sm.zpop_max(key, count))

    def zpop_min(
            self,
            key: str,
            count: int = 1,
            ) -> None:
        self.add_cmd(lambda: self._sm.zpop_min(key, count))

    def zrange(self, key: str, start: int, stop: int) -> None:
        self.add_cmd(lambda: self._sm.zrange(key, start, stop))

    def zcard(self, key: str) -> None:
        self.add_cmd(lambda: self._sm.zcard(key))

    def hset(self, key: str, mapping: dict[str, str]) -> None:
        self.add_cmd(lambda: self._sm.hset(key, mapping))

    def hdel(self, key: str, *fields: str) -> None:
        self.add_cmd(lambda: self._sm.hdel(key, *fields))

    def hget(self, key: str, field: str) -> None:
        self.add_cmd(lambda: self._sm.hget(key, field))

    def hmget(self, key: str, *fields: str) -> None:
        self.add_cmd(lambda: self._sm.hmget(key, *fields))

    def hincrby(self, key: str, field: str, inc: float | int) -> None:
        self.add_cmd(lambda: self._sm.hincrby(key, field, inc))

    def hkeys(self, key: str) -> None:
        self.add_cmd(lambda: self._sm.hkeys(key))

    def hvals(self, key: str) -> None:
        self.add_cmd(lambda: self._sm.hvals(key))

    def hgetall(self, key: str) -> None:
        self.add_cmd(lambda: self._sm.hgetall(key))



"""Defines the script backend for the memory runtime."""
import json
from collections.abc import Callable
from typing import Any, cast, Literal, TYPE_CHECKING, TypedDict

from redipy.api import PipelineAPI, RedisAPI
from redipy.backend.backend import Backend, ExecFunction
from redipy.graph.cmd import CommandObj
from redipy.graph.expr import BinOps, ExprObj, JSONType
from redipy.graph.seq import SequenceObj
from redipy.util import json_compact


if TYPE_CHECKING:
    from redipy.memory.rt import LocalRuntime
    from redipy.memory.state import Machine


ExecState = tuple[
    list[str],  # keyv
    list[JSONType],  # argv
    list[dict[str, JSONType]],  # arg names
    list[dict[str, str]],  # key names
    list[list[JSONType]],  # stack
    'LocalRuntime',  # fns
    'Machine',  # redis state
    list[JSONType],  # return value
    dict[str, str],  # keys
    dict[str, JSONType],  # args
]
"""The execution state of the script. For fast access this is a tuple instead
of a dictionary. Use the `STATE_*` constants to access the individual fields
by copying the value into a local variable first (local variables are index
lookups in python while global variables are dictionary lookups)."""


STATE_KEYV: Literal[0] = 0
"""The list of key arguments to the script."""
STATE_ARGV: Literal[1] = 1
"""The list of value arguments to the script."""
STATE_ARG_NAMES: Literal[2] = 2
"""Mapping of script argument names to their values."""
STATE_KEY_NAMES: Literal[3] = 3
"""Mapping of script key argument names to their values."""
STATE_STACK: Literal[4] = 4
"""The local variable stack of the script. Each stack frame is a list for fast
variable access via index. The last stack frame in the list is the current one.
"""
STATE_FNS: Literal[5] = 5
"""An interface for function calls."""
STATE_SM: Literal[6] = 6
"""An interface for raw memory access (redis-like memory)."""
STATE_RETURN: Literal[7] = 7
"""The return value stack. The rightmost value is the current return value of
the script or function."""
STATE_KEYS: Literal[8] = 8
"""The actual key arguments to the script. This value is filled before the
script starts executing."""
STATE_ARGS: Literal[9] = 9
"""The actual arguments to the script. This value is filled before the
script starts executing."""


Cmd = Callable[[ExecState], None]
"""A command callback executing a statement."""
ExprCmd = Callable[[ExecState], JSONType]
"""A command callback executing an expression."""


CmdContext = TypedDict('CmdContext', {
    "local_count": int,
    "local_names": dict[str, int],
})
"""Command context to keep track of local variables."""


class _Uninit():
    """An uninitialized variable. When a stack frame is created during
    execution values are initialized with this singleton value."""
    def __str__(self) -> str:
        return "UNINIT"

    def __repr__(self) -> str:
        return self.__str__()


UNINIT = _Uninit()
"""Marks an uninitialized variable."""


class LocalBackend(
        Backend[Cmd, ExprCmd, CmdContext, CmdContext, 'LocalRuntime']):
    """The script backend for the memory runtime."""
    def create_command_context(self) -> CmdContext:
        return {
            "local_count": 0,
            "local_names": {},
        }

    def finish(self, ctx: CmdContext, script: Cmd) -> Cmd:
        # state_keyv = STATE_KEYV
        # state_argv = STATE_ARGV
        state_arg_names = STATE_ARG_NAMES
        state_key_names = STATE_KEY_NAMES
        state_stack = STATE_STACK
        # state_fns = STATE_FNS
        # state_sm = STATE_SM
        # state_return = STATE_RETURN
        # state_keys = STATE_KEYS
        # state_args = STATE_ARGS

        frame_size = ctx["local_count"]

        def exec_script(state: ExecState) -> None:
            state[state_stack].append([cast(JSONType, UNINIT)] * frame_size)
            state[state_arg_names].append({})
            state[state_key_names].append({})
            script(state)

        return exec_script

    def compile_sequence(
            self, ctx: CmdContext, seq: SequenceObj) -> Cmd:
        state_keyv = STATE_KEYV
        state_argv = STATE_ARGV
        # state_arg_names = STATE_ARG_NAMES
        # state_key_names = STATE_KEY_NAMES
        # state_stack = STATE_STACK
        # state_fns = STATE_FNS
        # state_sm = STATE_SM
        # state_return = STATE_RETURN
        state_keys = STATE_KEYS
        state_args = STATE_ARGS

        cmds = [
            self.compile_command(ctx, cmd)
            for cmd in seq["cmds"]
        ]
        if seq["kind"] == "seq":

            def exec_cmds(state: ExecState) -> None:
                for cmd in cmds:
                    cmd(state)

            return exec_cmds

        if seq["kind"] == "function":
            raise ValueError("not implemented yet")

        key_order = seq["keyv"]
        arg_order = seq["argv"]

        def exec_frame(state: ExecState) -> None:
            keys = state[state_keys]
            assert not state[state_keyv]
            state[state_keyv].extend(keys[key] for key in key_order)
            args = state[state_args]
            assert not state[state_argv]
            state[state_argv].extend(args[arg] for arg in arg_order)
            for cmd in cmds:
                cmd(state)

        return exec_frame

    def compile_command(
            self, ctx: CmdContext, cmd: CommandObj) -> Cmd:
        # state_keyv = STATE_KEYV
        # state_argv = STATE_ARGV
        state_arg_names = STATE_ARG_NAMES
        state_key_names = STATE_KEY_NAMES
        state_stack = STATE_STACK
        # state_fns = STATE_FNS
        # state_sm = STATE_SM
        state_return = STATE_RETURN
        # state_keys = STATE_KEYS
        # state_args = STATE_ARGS

        def declare_var(name: str) -> int:
            ix = ctx["local_count"]
            ctx["local_count"] += 1
            ctx["local_names"][name] = ix
            return ix

        if cmd["kind"] == "assign" or cmd["kind"] == "declare":
            var_name = cmd["assign"]["name"]
            is_declare = cmd["kind"] == "declare"
            rhs = self.compile_expr(ctx, cmd["value"])

            if cmd["assign"]["kind"] in ("var", "index"):
                if is_declare:
                    cur_ix = declare_var(var_name)
                else:
                    cur_ix = ctx["local_names"][var_name]

                def exec_var_assign(state: ExecState) -> None:
                    state[state_stack][-1][cur_ix] = rhs(state)

                return exec_var_assign

            if cmd["assign"]["kind"] == "arg":

                def exec_arg_assign(state: ExecState) -> None:
                    state[state_arg_names][-1][var_name] = rhs(state)

                return exec_arg_assign

            if cmd["assign"]["kind"] == "key":

                def exec_key_assign(state: ExecState) -> None:
                    key = rhs(state)
                    state[state_key_names][-1][var_name] = f"{key}"

                return exec_key_assign

            raise ValueError(f"unknown kind {cmd['assign']['kind']}")
        if cmd["kind"] == "assign_at":
            var_name = cmd["assign"]["name"]
            arr_index = self.compile_expr(ctx, cmd["index"])
            rhs = self.compile_expr(ctx, cmd["value"])

            if cmd["assign"]["kind"] == "var":
                cur_ix = ctx["local_names"][var_name]

                def exec_var_assign_at(state: ExecState) -> None:
                    arr = cast(list, state[state_stack][-1][cur_ix])
                    ix = int(cast(int, arr_index(state)))
                    elem = rhs(state)
                    if ix == len(arr):
                        arr.append(elem)
                    else:
                        arr[ix] = elem

                return exec_var_assign_at

            raise ValueError(
                f"cannot assign to position of {cmd['assign']['kind']}")
        if cmd["kind"] == "stmt":
            stmt = self.compile_expr(ctx, cmd["expr"])

            def exec_stmt(state: ExecState) -> None:
                stmt(state)

            return exec_stmt
        if cmd["kind"] == "branch":
            condition = self.compile_expr(ctx, cmd['condition'])
            exec_then = self.compile_sequence(ctx, cmd["then"])
            exec_else = self.compile_sequence(ctx, cmd["else"])

            def exec_if(state: ExecState) -> None:
                if condition(state):
                    exec_then(state)
                else:
                    exec_else(state)

            return exec_if
        if cmd["kind"] == "for":
            assert cmd["index"]["kind"] == "index"
            assert cmd["value"]["kind"] == "var"
            index_ix = declare_var(cmd["index"]["name"])
            value_ix = declare_var(cmd["value"]["name"])
            array_val = self.compile_expr(ctx, cmd["array"])
            loop_body = self.compile_sequence(ctx, cmd["body"])

            def exec_for(state: ExecState) -> None:
                arr = array_val(state)
                for ix, val in enumerate(cast(list, arr)):
                    state[state_stack][-1][index_ix] = ix
                    state[state_stack][-1][value_ix] = val
                    loop_body(state)

            return exec_for
        if cmd["kind"] == "while":
            condition = self.compile_expr(ctx, cmd['condition'])
            loop_body = self.compile_sequence(ctx, cmd["body"])

            def exec_while(state: ExecState) -> None:
                while condition(state):
                    loop_body(state)

            return exec_while
        if cmd["kind"] == "return":
            if cmd["value"] is None:

                def exec_return_none(state: ExecState) -> None:
                    state[state_return].append(None)

                return exec_return_none
            ret_val = self.compile_expr(ctx, cmd["value"])

            def exec_return_val(state: ExecState) -> None:
                state[state_return].append(ret_val(state))

            return exec_return_val
        raise ValueError(f"unknown kind {cmd['kind']} for command {cmd}")

    def compile_expr(self, ctx: CmdContext, expr: ExprObj) -> ExprCmd:
        state_keyv = STATE_KEYV
        state_argv = STATE_ARGV
        state_arg_names = STATE_ARG_NAMES
        state_key_names = STATE_KEY_NAMES
        state_stack = STATE_STACK
        state_fns = STATE_FNS
        state_sm = STATE_SM
        # state_return = STATE_RETURN
        # state_keys = STATE_KEYS
        # state_args = STATE_ARGS

        if expr["kind"] == "var":
            var_name = expr["name"]
            var_ix = ctx["local_names"][var_name]

            def get_var(state: ExecState) -> JSONType:
                res = state[state_stack][-1][var_ix]
                if res is UNINIT:
                    raise ValueError(f"{var_name} is uninitialized")
                return res

            return get_var
        if expr["kind"] == "index":
            ix_name = expr["name"]
            ix_ix = ctx["local_names"][ix_name]

            def get_index(state: ExecState) -> JSONType:
                res = state[state_stack][-1][ix_ix]
                if res is UNINIT:
                    raise ValueError(f"{ix_name} is uninitialized")
                return int(cast(int, res))

            return get_index
        if expr["kind"] == "arg":
            arg_name = expr["name"]
            arg_readable = expr["readable"]

            def get_arg(state: ExecState) -> JSONType:
                try:
                    return state[state_arg_names][-1][arg_name]
                except IndexError as ie:
                    raise KeyError(f"unknown argument {arg_readable}") from ie

            return get_arg
        if expr["kind"] == "key":
            key_name = expr["name"]
            key_readable = expr["readable"]

            def get_key(state: ExecState) -> JSONType:
                try:
                    return state[state_key_names][-1][key_name]
                except IndexError as ie:
                    raise KeyError(f"unknown argument {key_readable}") from ie

            return get_key
        if expr["kind"] == "load_json_arg":
            arg_ix = expr["index"]
            return lambda state: state[state_argv][arg_ix]
        if expr["kind"] == "load_key_arg":
            key_ix = expr["index"]
            return lambda state: state[state_keyv][key_ix]
        if expr["kind"] == "val":
            val_type = expr["type"]
            value = expr["value"]
            if val_type == "list":
                res_json = json_compact(value)
                return lambda state: json.loads(res_json)
            if val_type == "bool":
                res: JSONType = bool(value)
            elif val_type == "int":
                res = int(cast(str, value))
            elif val_type == "float":
                res = float(cast(str, value))
            elif val_type == "str":
                res = f"{value}"
            elif val_type == "none":
                res = None
            else:
                raise ValueError(f"unknown value type {val_type} for {expr}")
            return lambda state: res
        if expr["kind"] == "unary":
            unary_op = self.compile_expr(ctx, expr["arg"])
            if expr["op"] == "not":
                return lambda state: not unary_op(state)
            raise ValueError(f"unknown op {expr['op']} for {expr}")
        if expr["kind"] == "binary":
            lhs = cast(
                Callable[[ExecState], Any],
                self.compile_expr(ctx, expr["left"]))
            rhs = cast(
                Callable[[ExecState], Any],
                self.compile_expr(ctx, expr["right"]))
            ops: dict[BinOps, ExprCmd] = {
                "and": lambda state: lhs(state) and rhs(state),
                "or": lambda state: lhs(state) or rhs(state),
                "add": lambda state: lhs(state) + rhs(state),
                "sub": lambda state: lhs(state) - rhs(state),
                "lt": lambda state: lhs(state) < rhs(state),
                "le": lambda state: lhs(state) <= rhs(state),
                "gt": lambda state: lhs(state) > rhs(state),
                "ge": lambda state: lhs(state) >= rhs(state),
                "eq": lambda state: lhs(state) == rhs(state),
                "ne": lambda state: lhs(state) != rhs(state),
            }
            op = ops.get(expr["op"])
            if op is not None:
                return op
            raise ValueError(f"unknown op {expr['op']} for {expr}")
        if expr["kind"] == "constant":
            raw = expr["raw"]

            def exec_const(state: ExecState) -> JSONType:
                return state[state_fns].get_constant(raw)

            return exec_const
        if expr["kind"] == "array_at":
            arr_ref = self.compile_expr(ctx, expr["var"])
            arr_ix = self.compile_expr(ctx, expr["index"])

            def exec_arr_at(state: ExecState) -> JSONType:
                return cast(list, arr_ref(state))[cast(int, arr_ix(state))]

            return exec_arr_at
        if expr["kind"] == "array_len":
            arr_ref = self.compile_expr(ctx, expr["var"])
            return lambda state: len(cast(list, arr_ref(state)))
        if expr["kind"] == "concat":
            exec_strs = [
                self.compile_expr(ctx, strobj)
                for strobj in expr["strings"]
            ]

            def exec_concat(state: ExecState) -> JSONType:
                strs = [f"{expr_str(state)}" for expr_str in exec_strs]
                return "".join(strs)

            return exec_concat
        if expr["kind"] == "call":
            exec_args = [
                self.compile_expr(ctx, arg_fn)
                for arg_fn in expr["args"]
            ]
            fn_name = expr["name"]

            def exec_call(state: ExecState) -> JSONType:
                args = [expr_arg(state) for expr_arg in exec_args]
                return state[state_fns].call_fn(state[state_sm], fn_name, args)

            return exec_call
        raise ValueError(f"unknown kind {expr['kind']} for expression {expr}")

    def create_executable(
            self,
            code: Cmd,
            runtime: 'LocalRuntime',
            ) -> ExecFunction:
        state_return = STATE_RETURN

        def exec_code_fn(
                *,
                keys: dict[str, str],
                args: dict[str, JSONType],
                active_rt: 'LocalRuntime',
                sm: 'Machine') -> JSONType:
            with active_rt.lock():
                success = False
                try:
                    state: ExecState = (
                        [],  # keyv
                        [],  # argv
                        [],  # arg names
                        [],  # key names
                        [],  # stack
                        active_rt,  # fns
                        sm,  # redis state
                        [],  # return value
                        keys,  # keys
                        args,  # args
                    )
                    code(state)
                    # print(f"state: {state}")
                    try:
                        res = state[state_return].pop()
                    except IndexError as e:
                        msg = "did you forget to call 'set_return_value'?"
                        raise ValueError(msg) from e
                    if res in ({}, []):
                        # NOTE: we turn empty lists or objects into None
                        # to have a consistent behavior with lua
                        res = None
                    success = True
                finally:
                    if not success:
                        print(f"state: {state}")
                return res

        def exec_code(
                *,
                keys: dict[str, str],
                args: dict[str, JSONType],
                client: RedisAPI | PipelineAPI | None = None) -> JSONType:
            active_rt = runtime
            if client is not None:
                from redipy.memory.rt import LocalPipeline, LocalRuntime

                if isinstance(client, LocalRuntime):
                    active_rt = client
                elif isinstance(client, LocalPipeline):
                    pipe = cast(LocalPipeline, client)
                    p_rt, p_sm = pipe.get_runtime_tuple()
                    pipe.add_cmd(
                        lambda: exec_code_fn(
                            keys=keys,
                            args=args,
                            active_rt=p_rt,
                            sm=p_sm))
                    return None
                else:
                    # FIXME: we could handle any runtime if we keep the
                    # intermediate representation around
                    get_rt = getattr(client, "get_memory_runtime")
                    if get_rt is None:
                        raise ValueError(f"incompatible runtime: {client}")
                    active_rt = get_rt()
            sm = active_rt.get_machine()
            return exec_code_fn(
                keys=keys, args=args, active_rt=active_rt, sm=sm)

        return exec_code



"""Tests utility functions."""
from typing import Any

from redipy.util import escape, indent, lua_fmt, retain_some, unescape


def test_retain_some() -> None:
    """Tests the function `retain_some`."""

    def test_rs(
            input_arr: list[int],
            count: int,
            output_arr: list[int],
            delete_arr: set[int],
            **kwargs: Any) -> None:
        res, to_delete = retain_some(
            input_arr, count, key=lambda v: v, **kwargs)
        assert res == output_arr
        assert set(to_delete) == delete_arr

    test_rs(
        [5, 3, 2, 1, 9, 3, 0, 6],
        4,
        [3, 3, 5, 6, 9],
        {0, 1, 2})
    test_rs(
        [5, 3, 2, 1, 9, 3, 0, 6],
        4,
        [3, 5, 6, 9],
        {0, 1, 2, 3},
        keep_last=False)
    test_rs(
        [5, 3, 2, 1, 9, 3, 0, 6],
        4,
        [6, 3, 2, 1, 0],
        {3, 5, 9},
        reverse=True)
    test_rs(
        [5, 3, 2, 1, 9, 3, 0, 6],
        4,
        [3, 2, 1, 0],
        {3, 5, 6, 9},
        reverse=True,
        keep_last=False)
    test_rs(
        [5, 3, 2, 1, 9, 3, 0, 6],
        5,
        [2, 3, 3, 5, 6, 9],
        {0, 1})
    test_rs(
        [5, 3, 2, 1, 9, 3, 0, 6],
        5,
        [3, 3, 5, 6, 9],
        {0, 1, 2},
        keep_last=False)
    test_rs(
        [5, 3, 2, 1, 9, 3, 0, 6],
        5,
        [6, 3, 3, 2, 1, 0],
        {5, 9},
        reverse=True)
    test_rs(
        [5, 3, 2, 1, 9, 3, 0, 6],
        5,
        [3, 3, 2, 1, 0],
        {5, 6, 9},
        reverse=True,
        keep_last=False)
    test_rs(
        [5, 3, 2, 1, 9, 3],
        5,
        [1, 2, 3, 3, 5, 9],
        set())
    test_rs(
        [5, 3, 2, 1, 9, 3],
        6,
        [1, 2, 3, 3, 5, 9],
        set(),
        keep_last=False)
    test_rs(
        [5, 3, 2],
        6,
        [2, 3, 5],
        set())
    test_rs(
        [5, 3, 2],
        6,
        [2, 3, 5],
        set(),
        keep_last=False)


def test_escape() -> None:
    """Tests the functions `escape` and `unescape`."""

    def test(text: str, subs: dict[str, str]) -> None:
        rsubs = {
            repl: key
            for key, repl in subs.items()
        }
        assert text == unescape(escape(text, subs), rsubs)

    test("abc", {"\n": "n"})
    test("abc\0\n", {"\n": "n"})
    test("\\n\n", {"\n": "n"})
    test("\\n0\\0\0\n", {"\n": "n"})

    test("abc", {"\0": "0"})
    test("abc\0\n", {"\0": "0"})
    test("\\n\n", {"\0": "0"})
    test("\\n0\\0\0\n", {"\0": "0"})

    test("abc", {"\n": "n", "\0": "0"})
    test("abc\0\n", {"\n": "n", "\0": "0"})
    test("\\n\n", {"\n": "n", "\0": "0"})
    test("\\n0\\0\0\n", {"\n": "n", "\0": "0"})


def test_indent() -> None:
    """Test indentation functions."""
    fmt_str = lua_fmt(r"""
        local res = {}
        local key = nil
        for ix, value in ipairs(arr) do
            if ix % 2 == 1 then
                key = value
            else
                res[#res + 1] = {key, value}
            end
        end
        return res
    """)
    ref = (
        r"local res = {}"
        "\n"
        r"local key = nil"
        "\n"
        r"for ix, value in ipairs(arr) do"
        "\n"
        r"  if ix % 2 == 1 then"
        "\n"
        r"    key = value"
        "\n"
        r"  else"
        "\n"
        r"    res[#res + 1] = {key, value}"
        "\n"
        r"  end"
        "\n"
        r"end"
        "\n"
        r"return res"
        "\n"
    )
    assert fmt_str == ref
    ref_indent = [
        r"  local res = {}",
        r"  local key = nil",
        r"  for ix, value in ipairs(arr) do",
        r"    if ix % 2 == 1 then",
        r"      key = value",
        r"    else",
        r"      res[#res + 1] = {key, value}",
        r"    end",
        r"  end",
        r"  return res",
    ]
    fmt_ind = indent(fmt_str, 2)
    assert fmt_ind == ref_indent



"""Tests a complex example class using redipy scripts."""
from collections.abc import Callable
from test.util import get_test_config
from typing import cast

import pytest

from redipy.api import RedisClientAPI
from redipy.backend.backend import ExecFunction
from redipy.graph.expr import JSONType
from redipy.main import Redis
from redipy.symbolic.expr import Strs
from redipy.symbolic.fun import ToIntStr, ToNum
from redipy.symbolic.rhash import RedisHash
from redipy.symbolic.rvar import RedisVar
from redipy.symbolic.seq import FnContext
from redipy.util import code_fmt, lua_fmt


GET_KEY_0 = "(redis.call(\"get\", key_0) or 0)"
RET = "return cjson.encode"
RC = "redis.call"
RP = "redipy.asintstr"
PLD = "redipy.pairlist_dict"
KEY_1_P = "(key_1) .. (\":\")"
EMPTY_OBJ = r"{}"


LUA_SET_VALUE = f"""
-- HELPERS START --
local redipy = {EMPTY_OBJ}
function redipy.asintstr (val)
    return math.floor(val)
end
-- HELPERS END --
--[[ KEYV
size
frame
]]
--[[ ARGV
field
value
]]
local key_0 = (KEYS[1])  -- size
local key_1 = (KEYS[2])  -- frame
local arg_0 = cjson.decode(ARGV[1])  -- field
local arg_1 = cjson.decode(ARGV[2])  -- value
{RC}("hset", (key_1) .. (":") .. ({RP}({GET_KEY_0})), arg_0, arg_1)
"""


LUA_GET_VALUE = f"""
-- HELPERS START --
local redipy = {EMPTY_OBJ}
function redipy.asintstr (val)
    return math.floor(val)
end
-- HELPERS END --
--[[ KEYV
size
frame
]]
--[[ ARGV
field
]]
local key_0 = (KEYS[1])  -- size
local key_1 = (KEYS[2])  -- frame
local arg_0 = cjson.decode(ARGV[1])  -- field
{RET}(({RC}("hget", (key_1) .. (":") .. ({RP}({GET_KEY_0})), arg_0) or nil))
"""


LUA_POP_FRAME = f"""
-- HELPERS START --
local redipy = {EMPTY_OBJ}
function redipy.asintstr (val)
    return math.floor(val)
end
function redipy.pairlist_dict (arr)
    local res = {EMPTY_OBJ}
    local key = nil
    for _, value in ipairs(arr) do
        if key ~= nil then
            res[key] = value
            key = nil
        else
            key = value
        end
    end
    return res
end
-- HELPERS END --
--[[ KEYV
size
frame
]]
--[[ ARGV
]]
local key_0 = (KEYS[1])  -- size
local key_1 = (KEYS[2])  -- frame
local var_0 = {PLD}({RC}("hgetall", {KEY_1_P} .. ({RP}({GET_KEY_0}))))
redis.call("del", {KEY_1_P} .. ({RP}({GET_KEY_0})))
if (tonumber({GET_KEY_0}) > 0) then
    redis.call("incrbyfloat", key_0, -1)
else
    redis.call("del", key_0)
end
return cjson.encode(var_0)
"""


LUA_GET_CASCADING = f"""
-- HELPERS START --
local redipy = {EMPTY_OBJ}
function redipy.asintstr (val)
    return math.floor(val)
end
-- HELPERS END --
--[[ KEYV
size
frame
]]
--[[ ARGV
field
]]
local key_0 = (KEYS[1])  -- size
local key_1 = (KEYS[2])  -- frame
local var_0 = key_1
local arg_0 = cjson.decode(ARGV[1])  -- field
local var_1 = tonumber({GET_KEY_0})
local var_2 = nil
local var_3 = nil
while ((var_2 == nil) and (var_1 >= 0)) do
    var_3 = (var_0) .. (":") .. (redipy.asintstr(var_1))
    var_2 = (redis.call("hget", var_3, arg_0) or nil)
    var_1 = (var_1 - 1)
end
return cjson.encode(var_2)
"""


class RStack:
    """An example class that simulates a key value stack."""
    def __init__(
            self,
            rt: RedisClientAPI,
            set_lua_script: Callable[[str | None, str | None], None]) -> None:
        self._rt = rt

        set_lua_script("set_value", LUA_SET_VALUE)
        self._set_value = self._set_value_script()
        set_lua_script("get_value", LUA_GET_VALUE)
        self._get_value = self._get_value_script()
        set_lua_script("pop_frame", LUA_POP_FRAME)
        self._pop_frame = self._pop_frame_script()
        set_lua_script("get_cascading", LUA_GET_CASCADING)
        self._get_cascading = self._get_cascading_script()
        set_lua_script(None, None)

    def key(self, base: str, name: str) -> str:
        """
        Compute the key.

        Args:
            base (str): The base key.

            name (str): The name.

        Returns:
            str: The key associated with the name.
        """
        return f"{base}:{name}"

    def push_frame(self, base: str) -> None:
        """
        Pushes a new stack frame.

        Args:
            base (str): The base key.
        """
        self._rt.incrby(self.key(base, "size"), 1)

    def pop_frame(self, base: str) -> dict[str, str]:
        """
        Pops the current stack frame and returns its values.

        Args:
            base (str): The base key.

        Returns:
            dict[str, str] | None: The content of the stack frame.
        """
        res = self._pop_frame(
            keys={
                "size": self.key(base, "size"),
                "frame": self.key(base, "frame"),
            },
            args={})
        if res is None:
            return {}
        return cast(dict, res)

    def set_value(self, base: str, field: str, value: str) -> None:
        """
        Set a value in the current stack frame.

        Args:
            base (str): The base key.

            field (str): The field.

            value (str): The value.
        """
        self._set_value(
            keys={
                "size": self.key(base, "size"),
                "frame": self.key(base, "frame"),
            },
            args={"field": field, "value": value})

    def get_value(self, base: str, field: str) -> JSONType:
        """
        Returns a value from the current stack frame.

        Args:
            base (str): The base key.

            field (str): The field.

        Returns:
            JSONType: The value.
        """
        return self._get_value(
            keys={
                "size": self.key(base, "size"),
                "frame": self.key(base, "frame"),
            },
            args={"field": field})

    def get_cascading(self, base: str, field: str) -> JSONType:
        """
        Returns a value from the stack. If the value is not in the current
        stack frame the value is recursively retrieved from the previous
        stack frames.

        Args:
            base (str): The base key.

            field (str): The field.

        Returns:
            JSONType: The value.
        """
        return self._get_cascading(
            keys={
                "size": self.key(base, "size"),
                "frame": self.key(base, "frame"),
            },
            args={"field": field})

    def _set_value_script(self) -> ExecFunction:
        ctx = FnContext()
        rsize = RedisVar(ctx.add_key("size"))
        rframe = RedisHash(Strs(
            ctx.add_key("frame"),
            ":",
            ToIntStr(rsize.get(default=0))))
        field = ctx.add_arg("field")
        value = ctx.add_arg("value")
        ctx.add(rframe.hset({
            field: value,
        }))
        ctx.set_return_value(None)
        return self._rt.register_script(ctx)

    def _get_value_script(self) -> ExecFunction:
        ctx = FnContext()
        rsize = RedisVar(ctx.add_key("size"))
        rframe = RedisHash(Strs(
            ctx.add_key("frame"),
            ":",
            ToIntStr(rsize.get(default=0))))
        field = ctx.add_arg("field")
        ctx.set_return_value(rframe.hget(field))
        return self._rt.register_script(ctx)

    def _pop_frame_script(self) -> ExecFunction:
        ctx = FnContext()
        rsize = RedisVar(ctx.add_key("size"))
        rframe = RedisHash(
            Strs(ctx.add_key("frame"), ":", ToIntStr(rsize.get(default=0))))
        lcl = ctx.add_local(rframe.hgetall())
        ctx.add(rframe.delete())

        b_then, b_else = ctx.if_(ToNum(rsize.get(default=0)).gt_(0))
        b_then.add(rsize.incrby(-1))
        b_else.add(rsize.delete())

        ctx.set_return_value(lcl)
        return self._rt.register_script(ctx)

    def _get_cascading_script(self) -> ExecFunction:
        ctx = FnContext()
        rsize = RedisVar(ctx.add_key("size"))
        base = ctx.add_local(ctx.add_key("frame"))
        field = ctx.add_arg("field")
        pos = ctx.add_local(ToNum(rsize.get(default=0)))
        res = ctx.add_local(None)
        cur = ctx.add_local(None)
        rframe = RedisHash(cur)

        loop = ctx.while_(res.eq_(None).and_(pos.ge_(0)))
        loop.add(cur.assign(Strs(base, ":", ToIntStr(pos))))
        loop.add(res.assign(rframe.hget(field)))
        loop.add(pos.assign(pos - 1))

        ctx.set_return_value(res)
        return self._rt.register_script(ctx)


@pytest.mark.parametrize("rt_lua", [False, True])
def test_stack(rt_lua: bool) -> None:
    """
    Tests a complex example class using redipy scripts.

    Args:
        rt_lua (bool): Whether to use the redis or memory runtime.
    """
    code_name = None
    lua_script = None

    def set_lua_script(name: str | None, lscript: str | None) -> None:
        nonlocal code_name
        nonlocal lua_script

        code_name = name
        lua_script = lscript

    def code_hook(code: list[str]) -> None:
        nonlocal code_name
        nonlocal lua_script

        if lua_script is None:
            return
        code_str = code_fmt(code)
        lscript = lua_fmt(lua_script)
        success = False
        try:
            assert code_str == lscript
            success = True
        finally:
            if not success:
                print(f"script name: {code_name}")

    redis = Redis(
        "redis" if rt_lua else "memory",
        cfg=get_test_config() if rt_lua else None,
        lua_code_hook=code_hook)
    stack = RStack(redis, set_lua_script)

    stack.set_value("foo", "a", "hi")
    assert stack.get_value("bar", "a") is None
    assert stack.get_value("foo", "a") == "hi"

    stack.set_value("foo", "b", "foo")
    stack.set_value("bar", "a", "bye")
    stack.set_value("bar", "b", "bar")

    assert stack.get_value("foo", "a") == "hi"
    assert stack.get_value("foo", "b") == "foo"
    assert stack.get_value("foo", "c") is None
    assert stack.get_value("foo", "d") is None

    assert stack.get_value("bar", "a") == "bye"
    assert stack.get_value("bar", "b") == "bar"
    assert stack.get_value("bar", "c") is None
    assert stack.get_value("bar", "d") is None

    stack.push_frame("foo")
    stack.set_value("foo", "b", "bb")
    stack.set_value("foo", "c", "cc")
    stack.set_value("foo", "d", "dd")

    stack.push_frame("bar")
    stack.set_value("bar", "a", "2a")
    stack.set_value("bar", "b", "2b")

    assert stack.get_value("foo", "a") is None
    assert stack.get_value("foo", "b") == "bb"
    assert stack.get_value("foo", "c") == "cc"
    assert stack.get_value("foo", "d") == "dd"

    assert stack.get_value("bar", "a") == "2a"
    assert stack.get_value("bar", "b") == "2b"
    assert stack.get_value("bar", "c") is None
    assert stack.get_value("bar", "d") is None

    assert stack.get_cascading("foo", "a") == "hi"
    assert stack.get_cascading("foo", "b") == "bb"
    assert stack.get_cascading("foo", "c") == "cc"
    assert stack.get_cascading("foo", "d") == "dd"

    assert stack.get_cascading("bar", "a") == "2a"
    assert stack.get_cascading("bar", "b") == "2b"
    assert stack.get_cascading("bar", "c") is None
    assert stack.get_cascading("bar", "d") is None

    stack.push_frame("foo")
    stack.set_value("foo", "b", "bbb")
    stack.set_value("foo", "c", "ccc")

    stack.push_frame("bar")
    stack.set_value("bar", "b", "3b")
    stack.set_value("bar", "c", "3c")

    assert stack.get_value("foo", "a") is None
    assert stack.get_value("foo", "b") == "bbb"
    assert stack.get_value("foo", "c") == "ccc"
    assert stack.get_value("foo", "d") is None

    assert stack.get_value("bar", "a") is None
    assert stack.get_value("bar", "b") == "3b"
    assert stack.get_value("bar", "c") == "3c"
    assert stack.get_value("bar", "d") is None

    assert stack.get_cascading("foo", "a") == "hi"
    assert stack.get_cascading("foo", "b") == "bbb"
    assert stack.get_cascading("foo", "c") == "ccc"
    assert stack.get_cascading("foo", "d") == "dd"

    assert stack.get_cascading("bar", "a") == "2a"
    assert stack.get_cascading("bar", "b") == "3b"
    assert stack.get_cascading("bar", "c") == "3c"
    assert stack.get_cascading("bar", "d") is None

    assert stack.pop_frame("foo") == {
        "b": "bbb",
        "c": "ccc",
    }
    assert stack.pop_frame("bar") == {
        "b": "3b",
        "c": "3c",
    }

    assert stack.get_value("foo", "a") is None
    assert stack.get_value("foo", "b") == "bb"
    assert stack.get_value("foo", "c") == "cc"
    assert stack.get_value("foo", "d") == "dd"

    assert stack.get_value("bar", "a") == "2a"
    assert stack.get_value("bar", "b") == "2b"
    assert stack.get_value("bar", "c") is None
    assert stack.get_value("bar", "d") is None

    assert stack.get_cascading("foo", "a") == "hi"
    assert stack.get_cascading("foo", "b") == "bb"
    assert stack.get_cascading("foo", "c") == "cc"
    assert stack.get_cascading("foo", "d") == "dd"

    assert stack.get_cascading("bar", "a") == "2a"
    assert stack.get_cascading("bar", "b") == "2b"
    assert stack.get_cascading("bar", "c") is None
    assert stack.get_cascading("bar", "d") is None

    assert stack.pop_frame("foo") == {
        "b": "bb",
        "c": "cc",
        "d": "dd",
    }
    assert stack.pop_frame("bar") == {
        "a": "2a",
        "b": "2b",
    }

    assert stack.get_value("foo", "a") == "hi"
    assert stack.get_value("foo", "b") == "foo"
    assert stack.get_value("foo", "c") is None
    assert stack.get_value("foo", "d") is None

    assert stack.get_value("bar", "a") == "bye"
    assert stack.get_value("bar", "b") == "bar"
    assert stack.get_value("bar", "c") is None
    assert stack.get_value("bar", "d") is None

    assert stack.get_cascading("foo", "a") == "hi"
    assert stack.get_cascading("foo", "b") == "foo"
    assert stack.get_cascading("foo", "c") is None
    assert stack.get_cascading("foo", "d") is None

    assert stack.get_cascading("bar", "a") == "bye"
    assert stack.get_cascading("bar", "b") == "bar"
    assert stack.get_cascading("bar", "c") is None
    assert stack.get_cascading("bar", "d") is None

    assert redis.exists("foo:size", "foo:frame:0") == 2
    assert redis.exists("foo:frame:1") == 0
    assert redis.exists("bar:size", "bar:frame:0") == 2
    assert redis.exists("bar:frame:1") == 0

    assert stack.pop_frame("foo") == {
        "a": "hi",
        "b": "foo",
    }
    assert stack.pop_frame("bar") == {
        "a": "bye",
        "b": "bar",
    }

    assert stack.get_cascading("foo", "a") is None
    assert stack.get_cascading("foo", "b") is None
    assert stack.get_cascading("foo", "c") is None
    assert stack.get_cascading("foo", "d") is None

    assert stack.get_cascading("bar", "a") is None
    assert stack.get_cascading("bar", "b") is None
    assert stack.get_cascading("bar", "c") is None
    assert stack.get_cascading("bar", "d") is None

    assert stack.pop_frame("foo") == {}
    assert stack.pop_frame("bar") == {}

    assert stack.get_cascading("foo", "a") is None
    assert stack.get_cascading("foo", "b") is None
    assert stack.get_cascading("foo", "c") is None
    assert stack.get_cascading("foo", "d") is None

    assert stack.get_cascading("bar", "a") is None
    assert stack.get_cascading("bar", "b") is None
    assert stack.get_cascading("bar", "c") is None
    assert stack.get_cascading("bar", "d") is None

    assert redis.exists("bar:size", "bar:frame:0") == 0
    assert redis.exists("bar:frame:1") == 0



"""Tests to verify unintuitive redis or lua behavior."""
from test.util import get_test_config

import pytest
import redis as redis_lib

from redipy.redis.conn import RedisConnection


def test_sanity() -> None:
    """Test to verify unintuitive redis or lua behavior."""
    redis = RedisConnection("test_sanity", cfg=get_test_config())

    def check_expression(
            raw: str,
            out: str,
            *,
            keys: list[str] | None = None,
            args: list[str] | None = None) -> None:
        if keys is None:
            keys = []
        if args is None:
            args = []
        code = f"return tostring({raw})"
        run = redis.get_dynamic_script(code)
        with redis.get_connection() as conn:
            res = run(
                keys=[redis.with_prefix(key) for key in keys],
                args=args,
                client=conn)
        assert res.decode("utf-8") == out

    # get
    check_expression("redis.call('get', KEYS[1])", "false", keys=["foo"])
    check_expression(
        "type(redis.call('get', KEYS[1]))", "boolean", keys=["foo"])
    assert redis.get("foo") is None

    # set
    check_expression(
        "cjson.encode(redis.call('set', KEYS[1], 'd'))",
        r'{"ok":"OK"}',
        keys=["bar"])
    check_expression(
        "cjson.encode(redis.call('set', KEYS[1], 'a', 'NX'))",
        "false",
        keys=["bar"])
    check_expression(
        "cjson.encode(redis.call('set', KEYS[1], 'a', 'XX'))",
        r'{"ok":"OK"}',
        keys=["bar"])
    check_expression("redis.call('get', KEYS[1])", "a", keys=["bar"])
    check_expression(
        "type(redis.call('get', KEYS[1]))", "string", keys=["bar"])
    check_expression(
        "cjson.encode(redis.call('set', KEYS[1], 'b'))",
        r'{"ok":"OK"}',
        keys=["bar"])
    check_expression("redis.call('get', KEYS[1])", "b", keys=["bar"])
    check_expression(
        "type(redis.call('get', KEYS[1]))", "string", keys=["bar"])
    assert redis.get("bar") == "b"
    assert redis.set("baz", "c") is True
    assert redis.get("baz") == "c"

    # lpop
    check_expression("redis.call('lpop', KEYS[1])", "false", keys=["foo"])
    check_expression(
        "type(redis.call('lpop', KEYS[1]))", "boolean", keys=["foo"])
    assert redis.lpop("foo") is None

    # rpop
    check_expression("redis.call('rpop', KEYS[1])", "false", keys=["foo"])
    check_expression(
        "type(redis.call('rpop', KEYS[1]))", "boolean", keys=["foo"])
    assert redis.rpop("foo") is None

    # llen
    check_expression("redis.call('llen', KEYS[1])", "0", keys=["foo"])
    check_expression(
        "type(redis.call('llen', KEYS[1]))", "number", keys=["foo"])
    assert redis.llen("foo") == 0

    # zpopmax
    check_expression(
        "cjson.encode(redis.call('zpopmax', KEYS[1]))", r"{}", keys=["foo"])
    assert redis.zpop_max("foo") == []  # pylint: disable=C1803
    check_expression("redis.call('zadd', KEYS[1], 2, 'a')", "1", keys=["zbar"])
    check_expression(
        "cjson.encode(redis.call('zpopmax', KEYS[1]))",
        "[\"a\",\"2\"]",
        keys=["zbar"])
    check_expression("redis.call('zadd', KEYS[1], 2, 'a')", "1", keys=["zbar"])
    check_expression(
        "cjson.encode(redis.call('zpopmax', KEYS[1], 1))",
        "[\"a\",\"2\"]",
        keys=["zbar"])
    check_expression("redis.call('zadd', KEYS[1], 2, 'a')", "1", keys=["zbar"])
    check_expression("redis.call('zadd', KEYS[1], 3, 'b')", "1", keys=["zbar"])
    check_expression(
        "cjson.encode(redis.call('zpopmax', KEYS[1], 2))",
        "[\"b\",\"3\",\"a\",\"2\"]",
        keys=["zbar"])
    check_expression("redis.call('zadd', KEYS[1], 2, 'a')", "1", keys=["zbar"])
    assert redis.zpop_max("zbar", 2) == [("a", 2)]
    check_expression("redis.call('zadd', KEYS[1], 2, 'a')", "1", keys=["zbar"])
    check_expression("redis.call('zadd', KEYS[1], 3, 'b')", "1", keys=["zbar"])
    assert redis.zpop_max("zbar", 2) == [("b", 3), ("a", 2)]

    # zpopmin
    check_expression(
        "cjson.encode(redis.call('zpopmin', KEYS[1]))", r"{}", keys=["foo"])
    assert redis.zpop_min("foo") == []  # pylint: disable=C1803
    check_expression("redis.call('zadd', KEYS[1], 2, 'a')", "1", keys=["zbar"])
    check_expression(
        "cjson.encode(redis.call('zpopmin', KEYS[1]))",
        "[\"a\",\"2\"]",
        keys=["zbar"])
    check_expression("redis.call('zadd', KEYS[1], 2, 'a')", "1", keys=["zbar"])
    check_expression(
        "cjson.encode(redis.call('zpopmin', KEYS[1], 1))",
        "[\"a\",\"2\"]",
        keys=["zbar"])
    check_expression("redis.call('zadd', KEYS[1], 2, 'a')", "1", keys=["zbar"])
    check_expression("redis.call('zadd', KEYS[1], 3, 'b')", "1", keys=["zbar"])
    check_expression(
        "cjson.encode(redis.call('zpopmin', KEYS[1], 2))",
        "[\"a\",\"2\",\"b\",\"3\"]",
        keys=["zbar"])
    check_expression("redis.call('zadd', KEYS[1], 2, 'a')", "1", keys=["zbar"])
    assert redis.zpop_min("zbar", 2) == [("a", 2)]
    check_expression("redis.call('zadd', KEYS[1], 2, 'a')", "1", keys=["zbar"])
    check_expression("redis.call('zadd', KEYS[1], 3, 'b')", "1", keys=["zbar"])
    assert redis.zpop_min("zbar", 2) == [("a", 2), ("b", 3)]


def test_ensure_name_available() -> None:
    """Verifies that new top level functions introduced in redipy do not exist
    already in redis or lua and would cause a name clash."""
    redis = RedisConnection(
        "test_ensure_name_available", cfg=get_test_config())

    def check_name(
            name: str,
            args: str,
            exc_type: type[BaseException],
            error_msg: str,
            error_chain: list[tuple[type[BaseException], str]]) -> None:
        code = f"return {name}({args})"
        run = redis.get_dynamic_script(code)
        with pytest.raises(exc_type, match=error_msg) as exc_info:
            with redis.get_connection() as conn:
                run(
                    keys=[],
                    args=[],
                    client=conn)
        cur_exc = exc_info.value
        for chain_type, chain_message in error_chain:
            next_exc = cur_exc.__cause__
            assert next_exc is not None
            assert isinstance(next_exc, chain_type)
            assert chain_message in f"{next_exc}"
            cur_exc = next_exc

    check_name(
        "asintstr",
        "3.2",
        ValueError,
        r"Error while executing script\nCode:\nreturn asintstr\(3\.2\)\n\n"
        r"Context:\n  \n  \n  \n> return asintstr\(3\.2\)",
        [
            (
                redis_lib.exceptions.ResponseError,
                r"user_script:1: Script attempted to access nonexistent "
                r"global variable 'asintstr'",
            ),
        ])



"""This module defines the base runtime for the different backends."""
import contextlib
import threading
from collections.abc import Callable, Iterator
from typing import Any, Generic, Self, TypeVar

from redipy.api import RedisClientAPI
from redipy.backend.backend import Backend, ExecFunction
from redipy.graph.seq import SequenceObj
from redipy.symbolic.seq import FnContext


T = TypeVar('T')


class Runtime(Generic[T], RedisClientAPI):
    """
    The base class for the different backends.
    """
    def __init__(self) -> None:
        self._backend: Backend[T, Any, Any, Any, Self] | None = None
        self._compile_hook: Callable[[SequenceObj], None] | None = None
        self._code_hook: Callable[[T], None] | None = None
        self._lock = threading.RLock()

    @contextlib.contextmanager
    def lock(self) -> Iterator[None]:
        """
        Enters this runtime's exclusive area. The lock is used to safely create
        a script backend.

        Yields:
            None: Inside the block the lock is held.
        """
        with self._lock:
            yield

    def set_compile_hook(
            self, hook: Callable[[SequenceObj], None] | None) -> None:
        """
        Sets the compile hook that is called after compiling a script.

        Args:
            hook (Callable[[SequenceObj], None] | None): A function that takes
            the root sequence object as argument.
        """
        self._compile_hook = hook

    def set_code_hook(self, hook: Callable[[T], None] | None) -> None:
        """
        Sets the code hook that is called after compiling a script.

        Args:
            hook (Callable[[T], None] | None): A function that takes the
            final internal representation of the script as argument.
        """
        self._code_hook = hook

    def register_script(self, ctx: FnContext) -> ExecFunction:
        compiled = ctx.compile()
        if self._compile_hook is not None:
            self._compile_hook(compiled)

        backend = self.get_backend()
        code = backend.translate(compiled)
        if self._code_hook is not None:
            self._code_hook(code)
        return backend.create_executable(code, self)

    def get_backend(self) -> Backend[T, Any, Any, Any, Self]:
        """
        Returns the script backend. Every runtime has exactly one script
        backend. The backend is lazily generated using create_backend.

        Returns:
            Backend[T, Any, Any, Any, Self]: _description_
        """
        if self._backend is None:
            with self.lock():
                if self._backend is None:
                    self._backend = self.create_backend()
        return self._backend

    @classmethod
    def create_backend(cls) -> Backend[T, Any, Any, Any, Self]:
        """
        Creates the script backend. Do not call this function directly.
        Call get_backend instead.

        Returns:
            Backend[T, Any, Any, Any, Self]: The new script backend.
        """
        raise NotImplementedError()



"""This module contains the main class for accessing redis. The Redis class
can be instantiated with different backends."""
import contextlib
import datetime
from collections.abc import Callable, Iterator
from typing import Literal, overload

from redipy.api import PipelineAPI, RedisClientAPI, RSetMode, RSM_ALWAYS
from redipy.backend.backend import ExecFunction
from redipy.backend.runtime import Runtime
from redipy.graph.seq import SequenceObj
from redipy.memory.rt import LocalRuntime
from redipy.redis.conn import RedisConfig, RedisConnection, RedisFactory
from redipy.symbolic.seq import FnContext


class Redis(RedisClientAPI):
    """
    This class is a wrapper around different runtime backends. Use this class
    to instantiate a redipy runtime.
    """
    def __init__(
            self,
            backend: Literal["memory", "redis", "custom", "infer"] = "infer",
            *,
            cfg: RedisConfig | None = None,
            host: str | None = None,
            port: int | None = None,
            passwd: str | None = None,
            redis_module: str | None = None,
            prefix: str | None = None,
            path: str | None = None,
            is_caching_enabled: bool = True,
            redis_factory: RedisFactory | None = None,
            rt: Runtime | None = None,
            lua_code_hook: Callable[[list[str]], None] | None = None,
            compile_hook: Callable[[SequenceObj], None] | None = None,
            verbose_lua_test: bool = False,
            ) -> None:
        """
        Creates a redis API. You can choose which backend runtime to use or
        you can provide your own. If called without arguments a memory runtime
        is initialized.

        Args:
            backend (Literal["memory", "redis", "custom", "infer"], optional):
            Explicitly states which backend to use. If omitted the backend is
            inferred from the rest of the arguments. Defaults to "infer".

            cfg (RedisConfig | None, optional): An object containing redis
            connection parameters. Only used for "redis" backend.
            Defaults to None.

            host (str | None, optional): The redis host. Only used for "redis"
            backend. Defaults to None.

            port (int | None, optional): The redis port. Only used for "redis"
            backend. Defaults to None.

            passwd (str | None, optional): The redis password. Only used for
            "redis" backend. Defaults to None.

            redis_module (str | None, optional): A prefix that is added to all
            keys in the redis backend. Only used for "redis" backend. Defaults
            to None.

            prefix (str | None, optional): A prefix that is added to all keys
            in the redis backend. This preceeds the redis_module. Only used for
            "redis" backend. Defaults to None.

            path (str | None, optional): The path to the redis configuration
            file and rdb file. The value is effectively ignored but can be
            accessed via the redis runtime. This can be useful when you want to
            control the redis server via python. Only used for "redis" backend.
            Defaults to None.

            is_caching_enabled (bool, optional): Whether redis connections are
            cached between command calls. Only used for "redis" backend.
            Defaults to True.

            redis_factory (RedisFactory | None, optional): An optional factory
            method to create the redis connection object. Only used for "redis"
            backend. Defaults to None.

            rt (Runtime | None, optional): The backend runtime object.
            Only used for "custom" backend. Defaults to None.

            lua_code_hook (Callable[[list[str]], None] | None, optional):
            A debugging hook that is called when registering a script with the
            generated lua code. Only used for "redis" backend.
            Defaults to None.

            compile_hook (Callable[[SequenceObj], None] | None, optional):
            A debugging hook that is called when registering a script with the
            internal execution graph as argument. This is available on every
            backend. Defaults to None.

            verbose_lua_test (bool, optional): A debugging feature that prints
            the lua code to standard out every time a script is registered.
            This also applies to manually registered scripts (as opposed to
            automatically generated lua scripts). Defaults to False.

        Raises:
            ValueError: If the wrong arguments are provided for a given
            backend. Note, that superfluous arguments are not detected. Only
            if a required argument is missing this exception is raised.
        """
        if backend == "infer":
            redis_cfgs = [cfg, host, port, passwd, prefix, path]
            if rt is not None:
                backend = "custom"
            elif any(val is not None for val in redis_cfgs):
                backend = "redis"
            else:
                backend = "memory"
        if backend == "custom":
            if rt is None:
                raise ValueError("rt must not be None for custom backend")
        elif backend == "redis" or (backend == "infer" and any(cfg)):
            if cfg is None:
                cfg = {
                    "host": "localhost" if host is None else host,
                    "port": 6379 if port is None else port,
                    "passwd": "" if passwd is None else passwd,
                    "prefix": "" if prefix is None else prefix,
                    "path": "." if path is None else path,
                }
            rrt = RedisConnection(
                "" if redis_module is None else redis_module,
                cfg=cfg,
                redis_factory=redis_factory,
                is_caching_enabled=is_caching_enabled,
                verbose_test=verbose_lua_test)
            if lua_code_hook is not None:
                rrt.set_code_hook(lua_code_hook)
            rt = rrt
        elif backend == "memory":
            rt = LocalRuntime()
        else:
            raise ValueError(f"unknown backend {backend}")
        self._rt: Runtime = rt
        rt.set_compile_hook(compile_hook)

    def get_runtime(self) -> Runtime:
        """
        Returns the associated runtime.

        Returns:
            Runtime: The current runtime.
        """
        return self._rt

    def maybe_get_redis_runtime(self) -> RedisConnection | None:
        """
        Returns the redis runtime if this is a redis connection.

        Returns:
            RedisConnection | None: The redis runtime if available.
            None otherwise.
        """
        if not isinstance(self._rt, RedisConnection):
            return None
        return self._rt

    def maybe_get_memory_runtime(self) -> LocalRuntime | None:
        """
        Returns the memory runtime if this is a memory connection.

        Returns:
            LocalRuntime | None: The memory runtime if available.
            None otherwise.
        """
        if not isinstance(self._rt, LocalRuntime):
            return None
        return self._rt

    def get_redis_runtime(self) -> RedisConnection:
        """
        Returns the redis runtime.

        Raises:
            ValueError: If this is not a redis connection.

        Returns:
            RedisConnection: The redis runtime.
        """
        res = self.maybe_get_redis_runtime()
        if res is None:
            raise ValueError("not a redis runtime")
        return res

    def get_memory_runtime(self) -> LocalRuntime:
        """
        Returns the memory runtime.

        Raises:
            ValueError: If this is not a memory connection.

        Returns:
            LocalRuntime: The memory runtime.
        """
        res = self.maybe_get_memory_runtime()
        if res is None:
            raise ValueError("not a memory runtime")
        return res

    def register_script(self, ctx: FnContext) -> ExecFunction:
        return self._rt.register_script(ctx)

    @contextlib.contextmanager
    def pipeline(self) -> Iterator[PipelineAPI]:
        with self._rt.pipeline() as pipe:
            yield pipe

    def exists(self, *keys: str) -> int:
        return self._rt.exists(*keys)

    def delete(self, *keys: str) -> int:
        return self._rt.delete(*keys)

    @overload
    def set(
            self,
            key: str,
            value: str,
            *,
            mode: RSetMode,
            return_previous: Literal[True],
            expire_timestamp: datetime.datetime | None,
            expire_in: float | None,
            keep_ttl: bool) -> str | None:
        ...

    @overload
    def set(
            self,
            key: str,
            value: str,
            *,
            mode: RSetMode,
            return_previous: Literal[False],
            expire_timestamp: datetime.datetime | None,
            expire_in: float | None,
            keep_ttl: bool) -> bool | None:
        ...

    @overload
    def set(
            self,
            key: str,
            value: str,
            *,
            mode: RSetMode = RSM_ALWAYS,
            return_previous: bool = False,
            expire_timestamp: datetime.datetime | None = None,
            expire_in: float | None = None,
            keep_ttl: bool = False) -> str | bool | None:
        ...

    def set(
            self,
            key: str,
            value: str,
            *,
            mode: RSetMode = RSM_ALWAYS,
            return_previous: bool = False,
            expire_timestamp: datetime.datetime | None = None,
            expire_in: float | None = None,
            keep_ttl: bool = False) -> str | bool | None:
        return self._rt.set(
            key,
            value,
            mode=mode,
            return_previous=return_previous,
            expire_timestamp=expire_timestamp,
            expire_in=expire_in,
            keep_ttl=keep_ttl)

    def get(self, key: str) -> str | None:
        return self._rt.get(key)

    def incrby(self, key: str, inc: float | int) -> float:
        return self._rt.incrby(key, inc)

    def lpush(self, key: str, *values: str) -> int:
        return self._rt.lpush(key, *values)

    def rpush(self, key: str, *values: str) -> int:
        return self._rt.rpush(key, *values)

    @overload
    def lpop(
            self,
            key: str,
            count: None = None) -> str | None:
        ...

    @overload
    def lpop(  # pylint: disable=signature-differs
            self,
            key: str,
            count: int) -> list[str] | None:
        ...

    def lpop(
            self,
            key: str,
            count: int | None = None) -> str | list[str] | None:
        return self._rt.lpop(key, count)

    @overload
    def rpop(
            self,
            key: str,
            count: None = None) -> str | None:
        ...

    @overload
    def rpop(  # pylint: disable=signature-differs
            self,
            key: str,
            count: int) -> list[str] | None:
        ...

    def rpop(
            self,
            key: str,
            count: int | None = None) -> str | list[str] | None:
        return self._rt.rpop(key, count)

    def lrange(self, key: str, start: int, stop: int) -> list[str]:
        return self._rt.lrange(key, start, stop)

    def llen(self, key: str) -> int:
        return self._rt.llen(key)

    def zadd(self, key: str, mapping: dict[str, float]) -> int:
        return self._rt.zadd(key, mapping)

    def zpop_max(
            self,
            key: str,
            count: int = 1,
            ) -> list[tuple[str, float]]:
        return self._rt.zpop_max(key, count)

    def zpop_min(
            self,
            key: str,
            count: int = 1,
            ) -> list[tuple[str, float]]:
        return self._rt.zpop_min(key, count)

    def zrange(self, key: str, start: int, stop: int) -> list[str]:
        return self._rt.zrange(key, start, stop)

    def zcard(self, key: str) -> int:
        return self._rt.zcard(key)

    def hset(self, key: str, mapping: dict[str, str]) -> int:
        return self._rt.hset(key, mapping)

    def hdel(self, key: str, *fields: str) -> int:
        return self._rt.hdel(key, *fields)

    def hget(self, key: str, field: str) -> str | None:
        return self._rt.hget(key, field)

    def hmget(self, key: str, *fields: str) -> dict[str, str | None]:
        return self._rt.hmget(key, *fields)

    def hincrby(self, key: str, field: str, inc: float | int) -> float:
        return self._rt.hincrby(key, field, inc)

    def hkeys(self, key: str) -> list[str]:
        return self._rt.hkeys(key)

    def hvals(self, key: str) -> list[str]:
        return self._rt.hvals(key)

    def hgetall(self, key: str) -> dict[str, str]:
        return self._rt.hgetall(key)



"""Access to redis variables."""
from redipy.api import RSetMode, RSM_ALWAYS, RSM_EXISTS, RSM_MISSING
from redipy.symbolic.expr import Expr, MixedType
from redipy.symbolic.fun import RedisObj


class RedisVar(RedisObj):
    """A redis variable."""
    def set(
            self,
            value: MixedType,
            *,
            mode: RSetMode = RSM_ALWAYS,
            return_previous: bool = False,
            expire_in: float | None = None,
            keep_ttl: bool = False) -> Expr:
        """
        Sets the value.

        Args:
            value (MixedType): The value.

            mode (RSetMode, optional): The condition to set the value. Defaults
            to RSM_ALWAYS.

            return_previous (bool, optional): Whether to return the previous
            value. Defaults to False.

            expire_in (float | None, optional): Expires the value in seconds.
            Defaults to None.

            keep_ttl (bool, optional): Preserve the time to live. Defaults to
            False.

        Returns:
            Expr: The expression.
        """
        args: list[MixedType] = []
        if mode == RSM_EXISTS:
            args.append("XX")
        elif mode == RSM_MISSING:
            args.append("NX")
        if return_previous:
            args.append("GET")
        if expire_in is not None:
            args.append("PX")
            expire_milli = int(expire_in * 1000.0)
            args.append(expire_milli)
        elif keep_ttl:
            args.append("KEEPTTL")
        return self.redis_fn("set", value, *args)

    def get(
            self,
            *,
            default: MixedType = None,
            no_adjust: bool = False) -> Expr:
        """
        Returns the value.

        Args:
            default (MixedType, optional): The default value to return if
            the key does not exist.

            no_adjust (bool, optional): Whether to prevent patching the
            function call. This should not be neccessary. Defaults to False.

        Returns:
            Expr: The expression.
        """
        if default is not None:
            return self.redis_fn("get", no_adjust=True).or_(default)
        return self.redis_fn("get", no_adjust=no_adjust)

    def incrby(self, inc: MixedType) -> Expr:
        """
        Updates the numeric value by a given amount.

        Args:
            inc (MixedType): The relative amount.

        Returns:
            Expr: The expression.
        """
        return self.redis_fn("incrby", inc)



"""Tests sorted sets."""
from test.util import get_setup, run_code

import pytest

from redipy.backend.backend import ExecFunction
from redipy.symbolic.fun import FindFn
from redipy.symbolic.rzset import RedisSortedSet
from redipy.symbolic.seq import FnContext
from redipy.util import lua_fmt


EMPTY_OBJ = r"{}"
PLS = "redipy.pairlist_scores"


LUA_SCRIPT = lua_fmt(f"""
-- HELPERS START --
local redipy = {EMPTY_OBJ}
function redipy.nil_or_index (val)
    if val ~= nil then
        val = val - 1
    end
    return val
end
function redipy.pairlist_scores (arr)
    local res = {EMPTY_OBJ}
    local key = nil
    for ix, value in ipairs(arr) do
        if ix % 2 == 1 then
            key = value
        else
            res[#res + 1] = {{key, tonumber(value)}}
        end
    end
    return res
end
-- HELPERS END --
--[[ KEYV
zset
]]
--[[ ARGV
prefix
]]
local arg_0 = cjson.decode(ARGV[1])  -- prefix
local key_0 = (KEYS[1])  -- zset
local var_0 = cjson.decode("[]")
for ix_0, val_0 in ipairs({PLS}(redis.call("zpopmin", key_0, 5))) do
    if (redipy.nil_or_index(string.find(val_0[0 + 1], arg_0)) == 0) then
        var_0[#var_0 + 1] = val_0[0 + 1]
    end
end
return cjson.encode(var_0)
""")


ASET: dict[str, float] = {
    "a_x_0": 0.0,
    "a_p_1": 1.0,
    "a_s_2": 2.0,
    "a_d_3": 3.0,
    "a_f_4": 4.0,
    "b_v_0": 0.5,
    "b_t_1": 1.0,
    "b_q_2": 1.5,
    "b_u_3": 2.0,
    "c_e_0": 2.0,
    "c_a_1": 2.1,
    "c_i_2": 2.2,
    "c_k_3": 2.3,
    "c_l_4": 2.4,
    "c_b_5": 2.5,
}


BSET: dict[str, float] = {
    "a_z_0": 0,
    "a_y_1": 1,
    "a_x_2": 2,
    "a_w_3": 3,
    "a_v_4": 4,
    "b_u_0": 5,
    "b_t_1": 6,
    "b_s_2": 7,
    "b_r_3": 8,
    "c_q_0": 9,
    "c_p_1": 10,
    "c_o_2": 11,
    "c_n_3": 12,
    "c_m_4": 13,
    "c_l_5": 14,
    "c_k_6": 15,
}


RUN_TESTS: list[tuple[tuple[str, str], list[str] | None]] = [
    # -- a --
    # a_x_0: 0.0
    # b_v_0: 0.5
    # a_p_1: 1.0
    # b_t_1: 1.0
    # b_q_2: 1.5
    (("a", "a_"), ["a_x_0", "a_p_1"]),
    # -- b --
    # a_z_0: 0
    # a_y_1: 1
    # a_x_2: 2
    # a_w_3: 3
    # a_v_4: 4
    (("b", "a_"), ["a_z_0", "a_y_1", "a_x_2", "a_w_3", "a_v_4"]),
    # -- a --
    # a_s_2: 2.0
    # b_u_3: 2.0
    # c_e_0: 2.0
    # c_a_1: 2.1
    # c_i_2: 2.2
    (("a", "b_"), ["b_u_3"]),
    # -- b --
    # b_u_0: 5
    # b_t_1: 6
    # b_s_2: 7
    # b_r_3: 8
    # c_q_0: 9
    (("b", "b_"), ["b_u_0", "b_t_1", "b_s_2", "b_r_3"]),
    # -- a --
    # c_k_3: 2.3
    # c_l_4: 2.4
    # c_b_5: 2.5
    # a_d_3: 3.0
    # a_f_4: 4.0
    (("a", "c_"), ["c_k_3", "c_l_4", "c_b_5"]),
    # -- b --
    # c_p_1: 10
    # c_o_2: 11
    # c_n_3: 12
    # c_m_4: 13
    # c_l_5: 14
    (("b", "c_"), ["c_p_1", "c_o_2", "c_n_3", "c_m_4", "c_l_5"]),
    # -- b --
    # c_k_6: 15
    (("b", "a_"), None),
]


@pytest.mark.parametrize("rt_lua", [False, True])
def test_rzset(rt_lua: bool) -> None:
    """
    Tests sorted sets.

    Args:
        rt_lua (bool): Whether to use the redis or memory runtime.
    """
    rt = get_setup("test_rzset", rt_lua, lua_script=LUA_SCRIPT)

    ctx = FnContext()
    prefix = ctx.add_arg("prefix")
    zset = RedisSortedSet(ctx.add_key("zset"))
    res = ctx.add_local([])
    loop, _, cur = ctx.for_(zset.pop_min(5))
    b_then, _ = loop.if_(FindFn(cur[0], prefix).eq_(0))
    b_then.add(res.set_at(res.len_(), cur[0]))
    ctx.set_return_value(res)

    rt.zadd("a", ASET)
    rt.zadd("b", BSET)
    rt.zadd("c", {"a": 0, "b": 1})

    def tester(
            runner: ExecFunction,
            vals: tuple[str, str]) -> list[str] | None:
        res = runner(keys={"zset": vals[0]}, args={"prefix": vals[1]})
        if res:
            assert isinstance(res, list)
        else:
            assert res is None
        return res

    run_code(rt, ctx, tests=RUN_TESTS, tester=tester)
    assert rt.zcard("a") == 0
    assert rt.zcard("b") == 0
    assert rt.zpop_max("c") == [("b", 1)]
    assert rt.zcard("c") == 1



"""Defines the script backend for the redis runtime. Scripts are converted to
lua."""
import json
from collections.abc import Iterable
from typing import Literal, TYPE_CHECKING

import redis as redis_lib

from redipy.api import PipelineAPI, RedisAPI
from redipy.backend.backend import Backend, ExecFunction
from redipy.graph.cmd import CommandObj
from redipy.graph.expr import BinOps, CallObj, ExprObj, get_literal, JSONType
from redipy.graph.seq import SequenceObj
from redipy.plugin import (
    add_patch_plugin,
    add_plugin,
    HELPER_PKG,
    HelperFunction,
    LuaGeneralPatch,
    LuaRedisPatch,
)
from redipy.util import code_fmt, indent, json_compact, lua_fmt


if TYPE_CHECKING:
    from redipy.redis.conn import RedisConnection


class LuaFnHook:
    """Hook for compiling general and redis function calls to lua."""
    def __init__(
            self,
            helper_fns: dict[str, HelperFunction],
            general_patch_fns: dict[str, LuaGeneralPatch],
            redis_patch_fns: dict[str, LuaRedisPatch]) -> None:
        """
        Creates a hook for compiling lua function calls.

        Args:
            helper_fns (dict[str, HelperFunction]): Helper function plugins.

            general_patch_fns (dict[str, LuaGeneralPatch]): General function
            patch plugins.

            redis_patch_fns (dict[str, LuaRedisPatch]): Redis function patch
            plugins.
        """
        self._helpers: set[str] = set()
        self._helper_fns = helper_fns
        self._general_patch_fns = general_patch_fns
        self._redis_patch_fns = redis_patch_fns
        self._is_expr_stmt = False

    def set_expr_stmt(self, is_expr_stmt: bool) -> None:
        """
        Sets whether we are currently handling the top expression of an
        expression statement.

        Args:
            is_expr_stmt (bool): Whether we are at the top expression of an
            expression statement.
        """
        self._is_expr_stmt = is_expr_stmt

    def is_expr_stmt(self) -> bool:
        """
        Whether the current expression is the top expression of an expression
        statement. If that is the case we cannot transform the expression into
        a side-effect free expression as lua will not allow the code to
        compile.

        Returns:
            bool: Whether the current expression is the top of an expression
            statement.
        """
        return self._is_expr_stmt

    def build_helpers(self) -> list[str]:
        """
        Compiles all helpers that were requested.

        Raises:
            RuntimeError: If a helper hasn't been registered.

        Returns:
            list[str]: All output code lines (without newline characters).
        """
        res = []
        prefix = f"{HELPER_PKG}."
        for helper in sorted(self._helpers):
            short_name = helper.removeprefix(prefix)
            help_obj = self._helper_fns.get(short_name)
            if help_obj is None:
                raise RuntimeError(f"unknown helper {short_name}")
            res.append(f"function {helper} ({help_obj.args()})")
            res.extend(indent(lua_fmt(help_obj.body()), 2))
            res.append("end")
        return res

    def adjust_function(self, expr: CallObj, is_expr_stmt: bool) -> ExprObj:
        """
        Patches up a function call by changing the execution graph. This is
        useful for when a function does not return what we want it to and we
        need to transform the function result. In order to not patch the
        function call the `"no_adjust"` field of the call graph object can be
        set to True. However, this should in general be avoided.

        Args:
            expr (CallObj): The expression to patch.

            is_expr_stmt (bool): Whether the expression is the top level
            expression of an expression statement. In most cases lua rejects
            the patched up version of the function call if it occurs as
            stand-alone statement. In that case it is better to not execute the
            patch and just call the function.

        Returns:
            ExprObj: The patched expression.
        """
        expr["no_adjust"] = True
        name = expr["name"]
        args = expr["args"]
        if name == "redis.call":
            r_name = get_literal(args[0], "str")
            if r_name is not None:
                return self.adjust_redis_fn(
                    expr, f"{r_name}", args[1:], is_expr_stmt=is_expr_stmt)
        patch_fn = self._general_patch_fns.get(name)
        if patch_fn is not None:
            return patch_fn.patch(expr, is_expr_stmt=is_expr_stmt)
        if name.startswith(f"{HELPER_PKG}."):
            self._helpers.add(name)
            return expr
        return expr

    def adjust_redis_fn(
            self,
            expr: CallObj,
            name: str,
            args: list[ExprObj],
            *,
            is_expr_stmt: bool) -> ExprObj:
        """
        Patches up a redis function call by changing the execution graph. This
        is useful for when a redis function does not return what we want it to
        and we need to transform the result.

        Args:
            expr (CallObj): The expression to patch.

            name (str): The name of the redis function.

            args (list[ExprObj]): The argument list of the redis function.

            is_expr_stmt (bool): Whether the expression is the top level
            expression of an expression statement. In most cases lua rejects
            the patched up version of the redis function call if it occurs as
            stand-alone statement. In that case it is better to not execute the
            patch and just call the redis function.

        Returns:
            ExprObj: The patched expression.
        """
        patch_fn = self._redis_patch_fns.get(name)
        if patch_fn is None:
            return expr
        return patch_fn.patch(name, expr, args, is_expr_stmt=is_expr_stmt)


def indent_str(code: Iterable[str], add_indent: int) -> list[str]:
    """
    Indents a stream of lines by a certain amount of spaces.

    Args:
        code (Iterable[str]): The stream of lines.
        add_indent (int): The indent to add.

    Returns:
        list[str]: The indented list of lines.
    """
    ind = " " * add_indent
    return [f"{ind}{exe}" for exe in code]


KEYV_HOOK = "--[[ KEYV"
"""Hook for detecting key argument names."""
ARGV_HOOK = "--[[ ARGV"
"""Hook for detecting value argument names."""
HOOK_END = "]]"
"""Hook for detecting end of name section."""


class LuaBackend(
        Backend[list[str], str, LuaFnHook, LuaFnHook, 'RedisConnection']):
    """Backend for compiling a script into lua code."""
    def __init__(self) -> None:
        super().__init__()
        self._helper_fns: dict[str, HelperFunction] = {}
        self._redis_patch_fns: dict[str, LuaRedisPatch] = {}
        self._general_patch_fns: dict[str, LuaGeneralPatch] = {}
        self.add_helper_function_plugin("redipy.redis.helpers")
        self.add_general_patch_plugin("redipy.redis.gpatch")
        self.add_redis_patch_plugin("redipy.redis.rpatch")

    def add_helper_function_plugin(self, module: str) -> None:
        """
        Add lua helper function plugins from the given module.

        Args:
            module (str): The module.
        """
        add_plugin(module, self._helper_fns, HelperFunction)

    def add_general_patch_plugin(self, module: str) -> None:
        """
        Add general function patch plugins from the given module.

        Args:
            module (str): The module.
        """
        add_patch_plugin(
            module,
            self._general_patch_fns,
            LuaGeneralPatch,
            disallowed={"redis.call"})

    def add_redis_patch_plugin(self, module: str) -> None:
        """
        Add redis function patch plugins from the given module.

        Args:
            module (str): The module.
        """
        add_patch_plugin(module, self._redis_patch_fns, LuaRedisPatch)

    def create_command_context(self) -> LuaFnHook:
        return LuaFnHook(
            self._helper_fns, self._general_patch_fns, self._redis_patch_fns)

    def finish(self, ctx: LuaFnHook, script: list[str]) -> list[str]:
        res = []
        helpers = ctx.build_helpers()
        if helpers:
            res.append("-- HELPERS START --")
            res.append(f"local {HELPER_PKG} = {{}}")
            res.extend(helpers)
            res.append("-- HELPERS END --")
        res.extend(script)
        return res

    def compile_sequence(self, ctx: LuaFnHook, seq: SequenceObj) -> list[str]:
        res = []
        if seq["kind"] == "script":
            key_order = seq["keyv"]
            arg_order = seq["argv"]
            res.append(KEYV_HOOK)
            for key in key_order:
                res.append(key)
            res.append(HOOK_END)
            res.append(ARGV_HOOK)
            for arg in arg_order:
                res.append(arg)
            res.append(HOOK_END)

        res.extend(indent_str(
            (
                exe
                for cmd in seq["cmds"]
                for exe in self.compile_command(ctx, cmd)
            ),
            0 if seq["kind"] == "script" else 2))
        return res

    def compile_command(self, ctx: LuaFnHook, cmd: CommandObj) -> list[str]:
        # FIXME: add debug context in comments
        if cmd["kind"] == "assign" or cmd["kind"] == "declare":
            is_declare = cmd["kind"] == "declare"
            decl = "local " if is_declare else ""
            rhs = self.compile_expr(ctx, cmd["value"])
            assign_obj = cmd["assign"]
            lcl_name = assign_obj["name"]
            if assign_obj["kind"] == "arg" or assign_obj["kind"] == "key":
                ext_name = f"  -- {assign_obj['readable']}"
            else:
                ext_name = ""
            return [f"{decl}{lcl_name} = {rhs}{ext_name}"]
        if cmd["kind"] == "assign_at":
            arr_ix = self.compile_expr(ctx, cmd["index"])
            rhs = self.compile_expr(ctx, cmd["value"])
            assign_obj = cmd["assign"]
            if assign_obj["kind"] == "var":
                lcl_name = assign_obj["name"]
                return [f"{lcl_name}[{arr_ix} + 1] = {rhs}"]
            raise ValueError(
                f"cannot assign to position of {cmd['assign']['kind']}")
        if cmd["kind"] == "stmt":
            ctx.set_expr_stmt(True)
            stmt = self.compile_expr(ctx, cmd["expr"])
            return [stmt]
        if cmd["kind"] == "branch":
            res = [f"if {self.compile_expr(ctx, cmd['condition'])} then"]
            res.extend(self.compile_sequence(ctx, cmd["then"]))
            elses = self.compile_sequence(ctx, cmd["else"])
            if elses:
                res.append("else")
                res.extend(elses)
            res.append("end")
            return res
        if cmd["kind"] == "for":
            assert cmd["index"]["kind"] == "index"
            assert cmd["value"]["kind"] == "var"
            res = [(
                f"for {cmd['index']['name']}, {cmd['value']['name']} "
                f"in ipairs({self.compile_expr(ctx, cmd['array'])}) do"
            )]
            res.extend(self.compile_sequence(ctx, cmd["body"]))
            res.append("end")
            return res
        if cmd["kind"] == "while":
            res = [f"while {self.compile_expr(ctx, cmd['condition'])} do"]
            res.extend(self.compile_sequence(ctx, cmd["body"]))
            res.append("end")
            return res
        if cmd["kind"] == "return":
            if cmd["value"] is None:
                return []
            return [
                f"return cjson.encode({self.compile_expr(ctx, cmd['value'])})",
            ]
        raise ValueError(f"unknown kind {cmd['kind']} for command {cmd}")

    def compile_expr(self, ctx: LuaFnHook, expr: ExprObj) -> str:
        is_expr_stmt = ctx.is_expr_stmt()
        if is_expr_stmt:
            ctx.set_expr_stmt(False)
        if expr["kind"] == "var":
            return expr["name"]
        if expr["kind"] == "arg":
            return expr["name"]
        if expr["kind"] == "key":
            return expr["name"]
        if expr["kind"] == "index":
            return f"({expr['name']} - 1)"
        if expr["kind"] == "load_json_arg":
            return f"cjson.decode(ARGV[{expr['index'] + 1}])"
        if expr["kind"] == "load_key_arg":
            return f"(KEYS[{expr['index'] + 1}])"
        if expr["kind"] == "val":
            val_type = expr["type"]
            value = expr["value"]
            if val_type == "bool":
                return f"{value}".lower()
            if val_type in ("int", "float"):
                return f"{value}"
            if val_type == "str":
                res = f"{value}"
                res = res.replace("\"", "\\\"").replace("\n", "\\n")
                return f"\"{res}\""
            if val_type == "list":
                res = json_compact(value).decode("utf-8")
                res = res.replace("\"", "\\\"").replace("\n", "\\n")
                return f"cjson.decode(\"{res}\")"
            if val_type == "none":
                return "nil"
            raise ValueError(f"unknown value type {val_type} for {expr}")
        if expr["kind"] == "unary":
            if expr["op"] == "not":
                return f"(not {self.compile_expr(ctx, expr['arg'])})"
            raise ValueError(f"unknown op {expr['op']} for {expr}")
        if expr["kind"] == "binary":
            ops: dict[BinOps, str] = {
                "and": "and",
                "or": "or",
                "add": "+",
                "sub": "-",
                "lt": "<",
                "le": "<=",
                "gt": ">",
                "ge": ">=",
                "eq": "==",
                "ne": "~=",
            }
            op = ops.get(expr["op"])
            if op is not None:
                return (
                    f"({self.compile_expr(ctx, expr['left'])} "
                    f"{op} "
                    f"{self.compile_expr(ctx, expr['right'])})")
            raise ValueError(f"unknown op {expr['op']} for {expr}")
        if expr["kind"] == "constant":
            return f"{expr['raw']}"
        if expr["kind"] == "array_at":
            return (
                f"{self.compile_expr(ctx, expr['var'])}"
                f"[{self.compile_expr(ctx, expr['index'])} + 1]")
        if expr["kind"] == "array_len":
            return f"#{self.compile_expr(ctx, expr['var'])}"
        if expr["kind"] == "concat":
            return " .. ".join(
                f"({self.compile_expr(ctx, strobj)})"
                for strobj in expr["strings"])
        if expr["kind"] == "call":
            if not expr["no_adjust"]:
                adj_expr = ctx.adjust_function(expr, is_expr_stmt)
                return self.compile_expr(ctx, adj_expr)
            argstr = ", ".join(
                self.compile_expr(ctx, arg) for arg in expr["args"])
            return f"{expr['name']}({argstr})"
        raise ValueError(f"unknown kind {expr['kind']} for expression {expr}")

    def create_executable(
            self,
            code: list[str],
            runtime: 'RedisConnection') -> ExecFunction:
        key_order = []
        arg_order = []
        mode: Literal["none", "keyv", "argv"] = "none"
        for line in code:
            ln = line.strip()
            if ln == HOOK_END:
                mode = "none"
                continue
            if ln == KEYV_HOOK:
                mode = "keyv"
                continue
            if ln == ARGV_HOOK:
                mode = "argv"
                continue
            if mode == "keyv":
                key_order.append(ln)
                continue
            if mode == "argv":
                arg_order.append(ln)
                continue
        compute = runtime.get_dynamic_script(code_fmt(code))

        def interpret_result(res: bytes) -> JSONType:
            if res is None:
                return None
            # NOTE: it is impossible to distinguish between {} and [] in lua
            if res == br"{}":
                return None
            return json.loads(res)

        def exec_code_fn(
                *,
                keys: dict[str, str],
                args: dict[str, JSONType],
                conn: redis_lib.Redis) -> bytes:
            return compute(
                keys=[runtime.with_prefix(keys[key]) for key in key_order],
                args=[json_compact(args[arg]) for arg in arg_order],
                client=conn)

        def exec_code(
                *,
                keys: dict[str, str],
                args: dict[str, JSONType],
                client: RedisAPI | PipelineAPI | None = None) -> JSONType:
            active_rt = runtime
            if client is not None:
                from redipy.redis.conn import (
                    PipelineConnection,
                    RedisConnection,
                )

                if isinstance(client, RedisConnection):
                    active_rt = client
                elif isinstance(client, PipelineConnection):
                    exec_code_fn(
                        keys=keys, args=args, conn=client.get_pipeline())
                    client.add_fixup(interpret_result)
                    return None
                else:
                    # FIXME: we could handle any runtime if we keep the
                    # intermediate representation around
                    get_rt = getattr(client, "get_redis_runtime")
                    if get_rt is None:
                        raise ValueError(f"incompatible runtime: {client}")
                    active_rt = get_rt()

            with active_rt.get_connection() as conn:
                res = exec_code_fn(keys=keys, args=args, conn=conn)
            return interpret_result(res)

        return exec_code



"""Tests miscellaneous redis functionality."""
import io
from contextlib import redirect_stdout
from test.util import get_setup

import pytest

from redipy.symbolic.fun import FromJSON, LogFn, ToJSON, ToStr, TypeStr
from redipy.symbolic.rzset import RedisSortedSet
from redipy.symbolic.seq import FnContext
from redipy.util import lua_fmt


MSG = (
    "this message is emitted by test/test_misc.py#test_misc. "
    "the output is not verified for the redis backend but if "
    "you see this message in the redis logs it means the test works."
)


CARD_CALL = "redis.call(\"zcard\", key_0)"


LUA_SCRIPT = lua_fmt(f"""
--[[ KEYV
in
]]
--[[ ARGV
]]
redis.log(redis.LOG_WARNING, "{MSG}")
local var_0 = tostring(5)
local var_1 = cjson.encode(var_0)
var_1 = cjson.decode(var_1)
local var_2 = true
if (not (type(var_1) == "string")) then
    var_2 = false
else
    if (var_0 ~= var_1) then
        var_2 = false
    end
end
local key_0 = (KEYS[1])  -- in
redis.call("zadd", key_0, 1, "a")
redis.call("zadd", key_0, 2, "b")
redis.call("zpopmax", key_0, 1)
if (not (({CARD_CALL} > 0) and ({CARD_CALL} < 2))) then
    var_2 = false
end
if (tostring(nil) ~= "nil") then
    var_2 = false
end
return cjson.encode(var_2)
""")


@pytest.mark.parametrize("rt_lua", [False, True])
def test_misc(rt_lua: bool) -> None:
    """
    Tests miscellaneous redis functionality.

    Args:
        rt_lua (bool): Whether to use the redis or memory runtime.
    """
    rt = get_setup(
        "test_misc", rt_lua, lua_script=LUA_SCRIPT, no_compile_hook=True)

    ctx = FnContext()
    ctx.add(LogFn("warning", MSG))
    lcl_str = ctx.add_local(ToStr(5))
    lcl_json = ctx.add_local(ToJSON(lcl_str))
    ctx.add(lcl_json.assign(FromJSON(lcl_json)))
    lcl_res = ctx.add_local(True)

    b_then, b_else = ctx.if_(TypeStr(lcl_json).eq_("string").not_())
    b_then.add(lcl_res.assign(False))
    s_then, _ = b_else.if_(lcl_str.ne_(lcl_json))
    s_then.add(lcl_res.assign(False))

    zset = RedisSortedSet(ctx.add_key("in"))
    ctx.add(zset.add(1, "a"))
    ctx.add(zset.add(2, "b"))
    ctx.add(zset.pop_max(1))
    z_then, _ = ctx.if_(zset.card().gt_(0).and_(zset.card().lt_(2)).not_())
    z_then.add(lcl_res.assign(False))

    n_then, _ = ctx.if_(ToStr(None).ne_("nil"))
    n_then.add(lcl_res.assign(False))

    ctx.set_return_value(lcl_res)

    exec_fun = rt.register_script(ctx)
    out = io.StringIO()
    with redirect_stdout(out):
        res = exec_fun(keys={"in": "foo"}, args={})
        assert res
    cmp = "" if rt_lua else f"WARNING: {MSG}\n"
    assert cmp in out.getvalue()

    with rt.pipeline() as pipe:
        pipe.lpush("bar", "a", "b", "c")
        assert pipe.execute() == [3]
        pipe.llen("bar")
        assert pipe.execute() == [3]



"""Patching general lua functions."""
from redipy.graph.expr import CallObj, ExprObj
from redipy.plugin import LuaGeneralPatch


class GStringFindPatch(LuaGeneralPatch):
    """Ensuring the index returned by the function is 0-based (python)."""
    @staticmethod
    def names() -> set[str]:
        return {"string.find"}

    def patch(
            self,
            expr: CallObj,
            *,
            is_expr_stmt: bool) -> ExprObj:
        if is_expr_stmt:
            return expr
        return {
            "kind": "call",
            "name": f"{self.helper_pkg()}.nil_or_index",
            "args": [expr],
            "no_adjust": False,
        }


class GAsIntStrPatch(LuaGeneralPatch):
    """Adds a function to convert a number into an integer string."""
    @staticmethod
    def names() -> set[str]:
        return {"asintstr"}

    def patch(
            self,
            expr: CallObj,
            *,
            is_expr_stmt: bool) -> ExprObj:
        return {
            "kind": "call",
            "name": f"{self.helper_pkg()}.asintstr",
            "args": expr["args"],
            "no_adjust": False,
        }



"""Provides core functionality of sequences."""
from redipy.graph.cmd import CommandObj, StmtObj
from redipy.graph.expr import IndexObj, RefIdObj, VarObj
from redipy.graph.seq import SequenceObj
from redipy.symbolic.core import (
    CmdHelper,
    Compilable,
    JSONArg,
    KeyVariable,
    LocalVariable,
    Variable,
)
from redipy.symbolic.expr import Expr, lit_helper, MixedType


class Sequence:
    """A sequence groups together statements."""
    def __init__(self, ctx: 'FnContext') -> None:
        """
        Creates a sequence. You probably shouldn't instantiate it yourself.

        Args:
            ctx (FnContext): The base script context.
        """
        self._ctx = ctx
        self._seq: list[Compilable] = []

    def compile(self) -> SequenceObj:
        """
        Compiles the sequence..

        Returns:
            SequenceObj: The SequenceObj sequence.
        """
        raise NotImplementedError()

    def add(
            self,
            term: Compilable | Expr) -> None:
        """
        Adds a new statement to the sequence.

        Args:
            term (Compilable | Expr): The statement. Expressions will
            automatically be converted appropriately.
        """
        if isinstance(term, Compilable):
            self._seq.append(term)
        else:
            stmt: StmtObj = {
                "kind": "stmt",
                "expr": term.compile(),
            }
            self._seq.append(CmdHelper(lambda: stmt))

    def is_empty(self) -> bool:
        """
        Whether the sequence is empty.

        Returns:
            bool: If True, the sequence has no elements in it.
        """
        return not self._seq

    def get_cmds(self) -> list[CommandObj]:
        """
        Returns all statements in this sequence.

        Returns:
            list[CommandObj]: The commands.
        """
        return [stmt.compile() for stmt in self._seq]

    def for_(self, array: Expr) -> tuple['Sequence', Variable, Variable]:
        """
        Creates a for loop.

        Returns:
            tuple[Sequence, Variable, Variable]: The first element is the
            sequence of the body of the for loop. The second element is the
            item index variable and the third is the item value variable.
        """
        loop = ForLoop(self._ctx, array)
        self.add(loop)
        return loop.get_loop(), loop.get_index(), loop.get_value()

    def if_(self, condition: MixedType) -> tuple['Sequence', 'Sequence']:
        """
        Creates an if branch.

        Args:
            condition (MixedType): The condition to continue the loop.

        Returns:
            tuple[Sequence, Sequence]: The first sequence is the body of the
            successful branch. The second sequence is the body of the
            unsuccessful branch.
        """
        branch = Branch(self._ctx, condition)
        self.add(branch)
        return branch.get_success(), branch.get_failure()

    def while_(self, condition: MixedType) -> 'Sequence':
        """
        Creates a while loop.

        Args:
            condition (MixedType): The condition.

        Returns:
            Sequence: The body of the loop. It is executed until `condition` is
            False.
        """
        loop = WhileLoop(self._ctx, condition)
        self.add(loop)
        return loop.get_loop()


class Seq(Sequence):
    """A standard sequence that does not add a stack frame."""
    def compile(self) -> SequenceObj:
        return {
            "kind": "seq",
            "cmds": self.get_cmds(),
        }


class FnContext(Sequence):
    """The base context for a script. Arguments, statements, and the return
    value are set through this class."""
    def __init__(self) -> None:
        """The base context for a script. Arguments, statements, and the return
        value are set through this class."""
        super().__init__(self)
        self._args: list[tuple[str, JSONArg]] = []
        self._keys: list[tuple[str, KeyVariable]] = []
        self._anames: set[str] = set()
        self._knames: set[str] = set()
        self._locals: list[LocalVariable] = []
        self._loops: int = 0

    def add_arg(self, name: str) -> JSONArg:
        """
        Adds a value / JSON argument with the given name.

        Args:
            name (str): The name.

        Raises:
            ValueError: If the name already exists.

        Returns:
            JSONArg: The argument reference.
        """
        if name in self._anames:
            raise ValueError(f"ambiguous arg name: {name}")
        arg = JSONArg(name)
        arg.set_index(len(self._args))
        self._args.append((name, arg))
        self._anames.add(name)
        self.add(CmdHelper(arg.get_declare))
        return arg

    def add_key(self, name: str) -> KeyVariable:
        """
        Adds a key argument with the given name.

        Args:
            name (str): The name.

        Raises:
            ValueError: If the name already exists.

        Returns:
            KeyVariable: The argument reference.
        """
        if name in self._knames:
            raise ValueError(f"ambiguous key name: {name}")
        key = KeyVariable(name)
        key.set_index(len(self._keys))
        self._keys.append((name, key))
        self._knames.add(name)
        self.add(CmdHelper(key.get_declare))
        return key

    def add_local(self, init: MixedType) -> LocalVariable:
        """
        Adds a local variable.

        Args:
            init (MixedType): The initial value of the variable.

        Returns:
            LocalVariable: The reference to the variable.
        """
        local = LocalVariable(init)
        local.set_index(len(self._locals))
        self._locals.append(local)
        self.add(CmdHelper(local.get_declare))
        return local

    def add_loop(self) -> int:
        """
        Adds a loop. For internal use only.

        Returns:
            int: The loop number.
        """
        loop = self._loops
        self._loops += 1
        return loop

    def set_return_value(self, value: MixedType) -> None:
        """
        Sets the return value of the script.

        Args:
            value (MixedType): The return value.
        """
        if value is None:
            self.add(CmdHelper(lambda: {
                "kind": "return",
                "value": None,
            }))
            return
        expr = lit_helper(value)
        self.add(CmdHelper(lambda: {
            "kind": "return",
            "value": expr.compile(),
        }))

    def compile(self) -> SequenceObj:
        return {
            "kind": "script",
            "cmds": self.get_cmds(),
            "argv": [arg for arg, _ in self._args],
            "keyv": [key for key, _ in self._keys],
        }


class Branch(Compilable):
    """An if / else block."""
    def __init__(self, ctx: FnContext, condition: MixedType) -> None:
        """
        Creates a branch.

        Args:
            ctx (FnContext): The script context.
            condition (MixedType): The branch condition.
        """
        self._condition = lit_helper(condition)
        self._success = Seq(ctx)
        self._failure = Seq(ctx)

    def get_success(self) -> Sequence:
        """
        The success branch.

        Returns:
            Sequence: The command sequence if the condition is successful.
        """
        return self._success

    def get_failure(self) -> Sequence:
        """
        The failure branch.

        Returns:
            Sequence: The command sequence if the condition failed.
        """
        return self._failure

    def compile(self) -> CommandObj:
        return {
            "kind": "branch",
            "condition": self._condition.compile(),
            "then": self._success.compile(),
            "else": self._failure.compile(),
        }


class IndexVariable(Variable):
    """A variable referencing the array index in a for loop."""
    def get_declare_rhs(self) -> Expr:
        raise RuntimeError("must be used in for loop")

    def prefix(self) -> str:
        return "ix"

    def get_index_ref(self) -> IndexObj:
        """
        Returns a reference to the index.

        Returns:
            IndexObj: The index reference.
        """
        return {
            "kind": "index",
            "name": f"{self.prefix()}_{self.get_index()}",
        }

    def get_ref(self) -> RefIdObj:
        return self.get_index_ref()


class ValueVariable(Variable):
    """A variable referencing the array element value in a for loop."""
    def get_declare_rhs(self) -> Expr:
        raise RuntimeError("must be used in for loop")

    def prefix(self) -> str:
        return "val"

    def get_var_ref(self) -> VarObj:
        """
        Returns a reference to the element value.

        Returns:
            VarObj: The element value reference.
        """
        return {
            "kind": "var",
            "name": f"{self.prefix()}_{self.get_index()}",
        }

    def get_ref(self) -> RefIdObj:
        return self.get_var_ref()


class ForLoop(Compilable):
    """A for loop over an array."""
    def __init__(self, ctx: FnContext, array: Expr) -> None:
        """
        Creates a for loop over the contents of an array.

        Args:
            ctx (FnContext): The script context.
            array (Expr): The array.
        """
        loop_ix = ctx.add_loop()
        self._ix = IndexVariable()
        self._ix.set_index(loop_ix)
        self._val = ValueVariable()
        self._val.set_index(loop_ix)
        self._loop = Seq(ctx)
        self._array = array

    def get_index(self) -> Variable:
        """
        The index variable.

        Returns:
            Variable: The index variable.
        """
        return self._ix

    def get_value(self) -> Variable:
        """
        The element value variable.

        Returns:
            Variable: The element value variable.
        """
        return self._val

    def get_loop(self) -> Sequence:
        """
        The loop block.

        Returns:
            Sequence: The loop block.
        """
        return self._loop

    def compile(self) -> CommandObj:
        return {
            "kind": "for",
            "array": self._array.compile(),
            "index": self._ix.get_index_ref(),
            "value": self._val.get_var_ref(),
            "body": self._loop.compile(),
        }


class WhileLoop(Compilable):
    """A while loop."""
    def __init__(self, ctx: FnContext, condition: MixedType) -> None:
        """
        Creates a loop the repeats while a condition is met.

        Args:
            ctx (FnContext): The script context.
            condition (MixedType): The condition.
        """
        self._condition = lit_helper(condition)
        self._loop = Seq(ctx)

    def get_loop(self) -> Sequence:
        """The loop body."""
        return self._loop

    def compile(self) -> CommandObj:
        return {
            "kind": "while",
            "condition": self._condition.compile(),
            "body": self._loop.compile(),
        }



"""Access to redis hash types."""
from redipy.symbolic.expr import Expr, MixedType
from redipy.symbolic.fun import RedisObj


class RedisHash(RedisObj):
    """A redis hash."""
    def hset(self, mapping: dict[MixedType, MixedType]) -> Expr:
        """
        Sets the fields to the provided values.

        Args:
            mapping (dict[MixedType, MixedType]): Mapping of field names and
            values.

        Returns:
            Expr: The expression.
        """
        args = []
        for key, value in mapping.items():
            args.append(key)
            args.append(value)
        return self.redis_fn("hset", *args)

    def hdel(self, *fields: MixedType) -> Expr:
        """
        Deletes fields.

        Args:
            *fields (MixedType): The fields to remove.

        Returns:
            Expr: The expression.
        """
        return self.redis_fn("hdel", *fields)

    def hget(self, field: MixedType) -> Expr:
        """
        Returns the value of a field.

        Args:
            field (MixedType): The field.

        Returns:
            Expr: The expression.
        """
        return self.redis_fn("hget", field)

    def hmget(self, *fields: MixedType) -> Expr:
        """
        Get multiple fields.

        Args:
            *fields (MixedType): The fields to remove.

        Returns:
            Expr: The expression.
        """
        return self.redis_fn("hmget", *fields)

    def hincrby(self, field: MixedType, inc: MixedType) -> Expr:
        """
        Updates the numerical value of a field.

        Args:
            field (MixedType): The field.
            inc (MixedType): The relative change.

        Returns:
            Expr: The expression.
        """
        return self.redis_fn("hincrby", field, inc)

    def hkeys(self) -> Expr:
        """
        Returns the keys of the hash.

        Returns:
            Expr: The expression.
        """
        return self.redis_fn("hkeys")

    def hvals(self) -> Expr:
        """
        Returns the values of the hash.

        Returns:
            Expr: The expression.
        """
        return self.redis_fn("hvals")

    def hgetall(self) -> Expr:
        """
        Returns the full hash.

        Returns:
            Expr: The expression.
        """
        return self.redis_fn("hgetall")



"""Implements general global functions for the memory backend scripts."""
import json
from typing import cast

from redipy.graph.expr import JSONType
from redipy.plugin import ArgcSpec, LocalGeneralFunction


class GStringFindFn(LocalGeneralFunction):
    """Implements finding a substring in a string."""
    @staticmethod
    def name() -> str:
        return "string.find"

    @staticmethod
    def argc() -> ArgcSpec:
        return {
            "count": 2,
            "at_most": 3,
        }

    @staticmethod
    def call(args: list[JSONType]) -> JSONType:
        found_ix = f"{args[0]}".find(
            f"{args[1]}",
            int(cast(int, args[2])) if len(args) > 2 else None)
        return None if found_ix < 0 else found_ix


class GCJSONDecodeFn(LocalGeneralFunction):
    """Implements decoding a JSON."""
    @staticmethod
    def name() -> str:
        return "cjson.decode"

    @staticmethod
    def argc() -> ArgcSpec:
        return {
            "count": 1,
        }

    @staticmethod
    def call(args: list[JSONType]) -> JSONType:
        return json.loads(f"{args[0]}")


class GCJSONEncodeFn(LocalGeneralFunction):
    """Implements encoding a JSON."""
    @staticmethod
    def name() -> str:
        return "cjson.encode"

    @staticmethod
    def argc() -> ArgcSpec:
        return {
            "count": 1,
        }

    @staticmethod
    def call(args: list[JSONType]) -> JSONType:
        return json.dumps(
            f"{args[0]}",
            sort_keys=True,
            indent=None,
            separators=(",", ":"))


class GToNumberFn(LocalGeneralFunction):
    """Implements converting strings into numbers."""
    @staticmethod
    def name() -> str:
        return "tonumber"

    @staticmethod
    def argc() -> ArgcSpec:
        return {
            "count": 1,
        }

    @staticmethod
    def call(args: list[JSONType]) -> JSONType:
        val = cast(str, args[0])
        try:
            return int(val)
        except (ValueError, TypeError):
            return float(val)


class GAsIntStrFn(LocalGeneralFunction):
    """Implements converting numbers into integer strings."""
    @staticmethod
    def name() -> str:
        return "asintstr"

    @staticmethod
    def argc() -> ArgcSpec:
        return {
            "count": 1,
        }

    @staticmethod
    def call(args: list[JSONType]) -> JSONType:
        val = cast(str, args[0])
        try:
            return int(val)
        except (ValueError, TypeError):
            return int(float(val))


class GToStringFn(LocalGeneralFunction):
    """Implements converting a value into a string."""
    @staticmethod
    def name() -> str:
        return "tostring"

    @staticmethod
    def argc() -> ArgcSpec:
        return {
            "count": 1,
        }

    @staticmethod
    def call(args: list[JSONType]) -> JSONType:
        oval = args[0]
        if isinstance(oval, bool):
            return f"{oval}".lower()
        # TODO dict, list
        if oval is None:
            return "nil"
        return f"{oval}"


class GTypeFn(LocalGeneralFunction):
    """Implements extracting the type of a value."""
    @staticmethod
    def name() -> str:
        return "type"

    @staticmethod
    def argc() -> ArgcSpec:
        return {
            "count": 1,
        }

    @staticmethod
    def call(args: list[JSONType]) -> JSONType:
        tmap: dict[type, str] = {
            bool: "boolean",
            dict: "table",
            float: "number",
            int: "number",
            list: "table",
            str: "string",
            type(None): "nil",
        }
        return tmap[type(args[0])]


class GRedisLogFn(LocalGeneralFunction):
    """Implements logging."""
    @staticmethod
    def name() -> str:
        return "redis.log"

    @staticmethod
    def argc() -> ArgcSpec:
        return {
            "count": 2,
        }

    @staticmethod
    def call(  # pylint: disable=useless-return
            args: list[JSONType]) -> JSONType:
        print(f"{args[0]}: {args[1]}")
        return None
